from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from concurrent.futures import as_completed
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Iterator
from typing import Sequence
from typing import TypeVar

import numpy as np
import pandas as pd

from .utils import MzMLResourceFile
from .utils import PeptideSettings

Result = TypeVar("Result")


def get_processpool(workers: int) -> ProcessPoolExecutor:
    return ProcessPoolExecutor(max_workers=workers)


def parallel_result(
    exe: list[Callable[[], Result]],
    workers: int = 4,
) -> Iterator[Result]:
    if not exe:
        return
    nworkers = min(len(exe), workers)
    if nworkers <= 1:
        for e in exe:
            yield e()
        return
    with get_processpool(workers) as executor:
        futures = [executor.submit(e) for e in exe]
        for future in as_completed(futures):
            yield future.result()


def apply_func(series: pd.Series, func: Callable[[Any], Result]) -> pd.Series:
    return series.apply(func)


def parallel_apply(
    series: pd.Series,
    func: Callable[[Any], Result],
    workers: int,
) -> pd.Series:
    from functools import partial

    chunk_num = max(len(series) // workers, 1)
    dfs: list[Callable[[], Result]]
    if chunk_num > 1:
        dfs = [
            partial(apply_func, sdf, func) for sdf in np.array_split(series, chunk_num)
        ]
    else:
        dfs = [partial(apply_func, series, func)]
    return pd.concat(
        [s for s in parallel_result(dfs, workers)],
        axis=0,
        ignore_index=False,
    )


@dataclass
class TaskBase:
    pepxml_df: pd.DataFrame
    mzml: MzMLResourceFile
    settings: PeptideSettings
    level: int = 0
    number_of_bg_processes: int = 1

    def mem(self) -> int:
        return self.pepxml_df.memory_usage(deep=True).sum()


class Task(TaskBase, ABC):
    @abstractmethod
    def task_run(self) -> pd.DataFrame:
        pass


def parallel_tasks(
    tasks: Sequence[Task],
    *,
    workers: int = 4,
) -> Iterator[tuple[pd.DataFrame, MzMLResourceFile]]:
    if not tasks:
        return

    if workers <= 1 or len(tasks) == 1:
        for task in tasks:
            yield task.task_run(), task.mzml
        return
    # ProcessPoolExecutor re-intialize loggers in the worker processes
    # with initialize function?
    nworkers = min(len(tasks), workers)

    with get_processpool(nworkers) as executor:
        futures = {executor.submit(t.task_run): t for t in tasks}
        for future in as_completed(futures):
            yield future.result(), futures[future].mzml



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path

import click

# from click_didyoumean import DYMGroup

HIDDEN = os.environ.get("TURNOVER_FULL", "0") != "1"
IsFile = click.Path(dir_okay=False, file_okay=True)

CONFIG = Path("~/.turnover.toml").expanduser()


@dataclass
class Config:
    logfile: str | None = None
    loglevel: str | None = None
    user_config: str | None = None


pass_config = click.make_pass_decorator(Config, ensure=True)


def update_mailhost(mailhost: str) -> None:
    from . import config

    config.MAIL_SERVER = mailhost


def update_config(filename: str | Path) -> None:
    from . import config
    import tomli as tomllib

    try:
        with open(filename, "rb") as fp:
            d = tomllib.load(fp)
    except Exception as e:
        raise click.BadParameter(
            f"can't read configuration file: {e}",
            param_hint="config",
        )

    for k, v in d.items():
        if k == "website":
            continue
        k = k.upper()
        if hasattr(config, k):
            setattr(config, k, v)
        else:
            click.secho(f"unknown configuration attribute {k}", fg="red")


epilog = click.style("turnover commands\n", fg="magenta")

epilog = f"""{epilog}

If no `--config` file is supplied `turnover` checks for a default configuration
file at `~/.turnover.toml` (see `init-config` to make one).
"""


@click.group(epilog=epilog)
@click.option(
    "-l",
    "--level",
    type=click.Choice(
        ["info", "debug", "warning", "error", "critical"],
        case_sensitive=False,
    ),
    help="log level",
)
@click.option(
    "-f",
    "--logfile",
    type=click.Path(file_okay=True, dir_okay=False),
    help="log file to write to (use '-' to log to stderr)",
)
@click.option(
    "-m",
    "--mailhost",
    help="where to send emails to",
    metavar="HOST",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="configuration file for turnover [.TOML format]",
)
@click.version_option()
@click.pass_context
def cli(
    ctx: click.Context,
    level: str | None,
    logfile: str | None = None,
    config: str | None = None,
    mailhost: str | None = None,
) -> None:
    from .logger import init_logger

    ctx.obj = Config(
        logfile=logfile,
        loglevel=level,
        user_config=config,
    )
    if level is None:
        level = "WARNING"

    if config is not None:
        update_config(config)
    else:
        if CONFIG.exists():
            update_config(CONFIG)
    if mailhost is not None:
        update_mailhost(mailhost)

    init_logger(level=level, logfile=logfile)



from __future__ import annotations

import os
import signal
import subprocess
import sys
import time
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from threading import Thread
from types import ModuleType
from typing import Any
from typing import Callable
from typing import TypedDict

import click
import tomli as tomllib
import tomli_w
from typing_extensions import NotRequired

from .background import Processing
from .cli import CONFIG

IS_WIN32 = sys.platform == "win32"

# def dict2conf(outfile: str, dictionary: dict[str, Any]) -> None:

#     with open(outfile, mode="wb") as fp:
#         tomli_w.dump(dictionary, fp)


# def dict2conffp(fp: IO, dictionary: dict[str, Any]) -> None:
#     for k, v in dictionary.items():
#         print(k, " = ", file=fp, end="")
#         pprint(v, stream=fp)


class PopenArgs(TypedDict):
    process_group: NotRequired[int | None]
    creationflags: int
    preexec_fn: Callable[[], None] | None


@dataclass
class Runner:
    name: str
    cmd: list[str]
    directory: str = "."
    env: dict[str, str] | None = None
    showcmd: bool = True
    shell: bool = False
    prevent_sig: bool = False  # prevent Cntrl-C from propagating to child process

    def getenv(self) -> dict[str, str] | None:
        if not self.env:
            return None
        return {**os.environ, **self.env}

    def start(self) -> subprocess.Popen[bytes]:
        if self.showcmd:
            click.secho(" ".join(str(s) for s in self.cmd), fg="blue")

        kwargs = PopenArgs(creationflags=0, preexec_fn=None)
        if self.prevent_sig:
            if sys.platform == "win32":
                kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP
            else:
                if sys.version_info >= (3, 11):
                    kwargs["process_group"] = 0
                else:
                    kwargs["preexec_fn"] = os.setpgrp
        return subprocess.Popen(  # type: ignore
            self.cmd,
            cwd=self.directory,
            env=self.getenv(),
            shell=self.shell,
            **kwargs,
        )


def browser(url: str = "http://127.0.0.1:8000", sleep: float = 5.0) -> Thread:
    import webbrowser

    def run() -> None:
        time.sleep(sleep)
        webbrowser.open_new_tab(url)

    tr = Thread(target=run)
    tr.daemon = True  # exit when main process exits
    tr.start()
    return tr


def has_package(package: str) -> bool:
    return get_package(package) is not None


def get_package(package: str) -> ModuleType | None:
    import importlib

    try:
        return importlib.import_module(package)

    except ModuleNotFoundError:
        return None


def has_website() -> bool:
    return has_package("protein_turnover_website")


def default_conf() -> dict[str, Any]:
    # from tempfile import gettempdir

    conf = {
        "MOUNTPOINTS": [
            ("~", "HOME"),
        ],
        "JOBSDIR": "~/turnover_jobs",
        "CACHEDIR": "~/turnover_cache",
        "WEBSITE_STATE": "single_user",
    }
    return conf


def instance_conf(config: Path | str, ns: str | None = None) -> dict[str, Any]:
    # """We *must* have flask in our environment by now"""
    # from flask import Config  # pylint: disable=import-error
    # conf = Config(".")
    # conf.from_pyfile(config)
    # return conf
    try:
        with open(config, "rb") as fp:
            ret = tomllib.load(fp)
            if ns is not None:
                ret2 = ret.get(ns, None)
                return ret2 or {}
            return ret
    except tomllib.TOMLDecodeError as e:
        raise click.BadParameter(f'Can\'t read configuration file "{config}"') from e


def dump_config(configfile: Path, with_website: bool):
    from . import config

    def mod2dict(module: ModuleType) -> dict[str, Any]:
        d: dict[str, Any] = {}
        for k in dir(module):
            if k.isupper():
                a = getattr(module, k)
                if a is None:
                    # a = ""
                    continue
                d[k] = a

        d = dict(sorted(d.items()))
        return d

    HEADER = b"""# turnover configuration file.
# Remove/comment out any values you don't change since these are defaults.
"""
    d = mod2dict(config)
    if with_website:
        c = get_package("protein_turnover_website.config")
        assert c is not None
        d["website"] = mod2dict(c)

    click.secho(f"writing {configfile}")
    with open(configfile, "wb") as fp:
        fp.write(HEADER)
        tomli_w.dump(d, fp)


def waitress(app: str, port: int) -> list[str]:
    return [
        sys.executable,
        "-m",
        "waitress",
        f"--listen=127.0.0.1:{port}",
        app + ":applications",
    ]


def webrunner(
    *,
    browse: bool,
    workers: int,
    web_config: str | Path | None = None,  # from --web-config
    gunicorn: bool = False,
    view_only: bool = True,
    configfile: str | None = None,  # turnover config file
    defaults: dict[str, Any] | None = None,  # CACHEDIR etc. from commanline
    port: int = 8000,
    server_options: tuple[str, ...] = (),  # extra commandline arguments after --
    flask_app: str = "protein_turnover_website.wsgi",
    page: str | None = None,
    email: bool = False,
) -> None:
    """Run full website."""
    from .config import MAIL_SERVER

    if not has_package("protein_turnover_website"):
        click.secho(
            "Please install protein_turnover_website [pip install protein-turnover-website]!",
            fg="red",
            err=True,
        )
        raise click.Abort()
    if gunicorn and not has_package("gunicorn"):
        click.secho(
            "Please install gunicorn [pip install gunicorn]!",
            fg="red",
            err=True,
        )
        raise click.Abort()
    if web_config is None:
        if CONFIG.exists():
            web_config = CONFIG
    fp = None
    web_conf = default_conf()
    if MAIL_SERVER and email:
        web_conf["MAIL_SERVER"] = MAIL_SERVER

    if web_config:
        # just need JOBSDIR
        web_conf.update(
            instance_conf(web_config, ns="website"),
        )
    # from commandline so... last
    if defaults is not None:
        web_conf.update(defaults)

    # need to read config file just for jobsdir
    jobsdir = Path(web_conf["JOBSDIR"]).expanduser()
    if not jobsdir.exists():
        jobsdir.mkdir(parents=True, exist_ok=True)

    if not view_only:
        email_args = [] if email and MAIL_SERVER else ["--no-email"]
        cfg = [f"--config={configfile}"] if configfile is not None else []
        background = Runner(
            "background",
            [
                sys.executable,
                "-m",
                "protein_turnover",
                *cfg,
                "--level=info",
                "background",
                f"--workers={workers}",
                *email_args,
                str(jobsdir),
            ],
            directory=".",
            prevent_sig=True,
        )
    Url = f"127.0.0.1:{port}"
    if page is not None:
        Url += f"/{page}"
    with TemporaryDirectory() as td:
        filename = Path(td) / "flask.cfg"
        with filename.open("wb") as fp:
            tomli_w.dump(web_conf, fp)  # type: ignore

        if not gunicorn:
            website = Runner(
                "flask",
                [
                    sys.executable,
                    "-m",
                    "flask",
                    "--app",
                    flask_app,
                    "run",
                    f"--port={port}",
                    *server_options,
                ],
                env={"TURNOVER_SETTINGS": str(filename), "FLASK_DEBUG": "0"},
                prevent_sig=True,
            )
        else:
            website = Runner(
                "gunicorn",
                [
                    sys.executable,
                    "-m",
                    "gunicorn",
                    f"--bind={Url}",
                    *server_options,
                    flask_app,
                ],
                env={"TURNOVER_SETTINGS": str(filename)},
                prevent_sig=True,
            )
        if view_only:
            procs = [website]
        else:
            procs = [background, website]

        processes = [(p.name, p.start()) for p in procs]

        if browse:
            browser(url=f"http://{Url}", sleep=3.0)

        worker = Processing(jobsdir)
        ninterrupts = 0
        prev = datetime.now()
        while True:
            try:
                time.sleep(10.0)
                failed = False
                for n, tr in processes:
                    try:
                        retcode = tr.wait(1.0)
                        if retcode != 0:
                            click.echo(f"process failed {n}")
                            failed = True
                    except subprocess.TimeoutExpired:
                        pass
                if failed:
                    raise KeyboardInterrupt()

            except KeyboardInterrupt:
                # too long between ^C
                now = datetime.now()
                if (
                    ninterrupts > 0
                    and not view_only
                    and (now - prev).total_seconds() > 5
                ):
                    ninterrupts = 0
                    prev = now
                    continue
                ninterrupts += 1
                if ninterrupts >= 2 or view_only:
                    for name, tr in processes:
                        click.secho(f"terminating... {name}", fg="blue")
                        tr.send_signal(
                            signal.SIGINT if not IS_WIN32 else signal.CTRL_C_EVENT
                        )

                    for name, tr in processes:
                        try:
                            tr.wait(timeout=4.0)
                        except (OSError, subprocess.TimeoutExpired):
                            pass
                    sys.exit(os.EX_OK)

                prev = now

                if not view_only and worker.is_processing():
                    click.secho(
                        "Warning! The background process is running a job!",
                        fg="yellow",
                        bold=True,
                    )
                click.secho("interrupt... ^C again to terminate")

        # os.system("stty sane")



from dataclasses import dataclass, fields, replace
from pathlib import Path
import click

dpi = 96


@dataclass
class TurnoverConfig:
    MIN_RT: int = 2
    ABUNDANCE_CUTOFF: float = 0.01
    # split jobs into this many spectra
    # set to zero to not split
    NSPECTRA: int = 10000
    # quadrature limit for estimating area under curve (unused now)
    # QUAD_LIMIT: int = 500
    # https://docs.python.org/3/library/logging.html#logrecord-attributes
    LOG_FORMAT: str = "%(levelname)s|[%(asctime)s]|%(process)d| %(message)s"

    # XCMS_STEP = 0.0
    PEPXML_CHUNKS: int = 1000

    # also group on retention_time_sec
    GROUP_RT: bool = False

    # FIGSIZE: tuple[float, float] = (1056.0 / dpi, 768.0 / dpi)

    MAIL_SUBJECT: str = "turnover pipeline"
    # default "from" sender
    MAIL_SENDER: str = "turnover-pipeline@uwa.edu.au"
    # set to None or 'none' to stop any emailing
    MAIL_SERVER: str = "mailhost"

    # e.g. "https://protein-turnover.plantenergy.org/inspect/{jobid}"
    INSPECT_URL: str | None = None
    # email template with {job:TurnoverJob, url:str}
    MAIL_TEXT: str = (
        """Protein Turnover Job "{job.job_name}" has <b>finished</b>!{url}."""
    )
    # set to True for production version (hides debug click commands)

    INTERPOLATE_INTENSITY: bool = True

    # Dinosaur
    JAVA_PATH: str | None = None  # "java"
    DINOSAUR_JAR: str | None = None


FIELDS = [f.name for f in fields(TurnoverConfig)]

config = TurnoverConfig()


def _update_config(config: TurnoverConfig, filename: str | Path) -> TurnoverConfig:

    import tomli as tomllib

    updates = {}
    try:
        with open(filename, "rb") as fp:
            d = tomllib.load(fp)
    except Exception as e:
        raise click.BadParameter(
            f"can't read configuration file: {e}",
            param_hint="config",
        )

    for k, v in d.items():
        if k == "website":
            continue
        k = k.upper()

        if k in FIELDS:
            v2 = getattr(config, k)
            if v != v2:
                updates[k] = v

        else:
            click.secho(f"unknown configuration attribute {k}", fg="red")

    if updates:
        return replace(config, **updates)
    return config


def get_config() -> TurnoverConfig:
    return config


def update_config(filename: str | Path) -> TurnoverConfig:
    global config
    config = _update_config(config, filename)
    return config



from __future__ import annotations

import click

from .cli import cli
from .cli import Config
from .cli import pass_config


@cli.command(name="background")
@click.option(
    "--workers",
    type=int,
    help="number of workers [default - half the number of cpus]",
)
@click.option(
    "--nice",
    default=0,
    type=click.IntRange(
        min=0,
    ),
    help="run turnover command at this nice level",
)
@click.option(
    "--sleep",
    default=60.0,
    help="sleep (seconds) between directory scans",
    show_default=True,
)
@click.option(
    "--run-config",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="configuration file to pass off to the background process",
)
@click.option(
    "--no-email",
    is_flag=True,
    help="don't send any emails",
    hidden=False,
)  # see runner.py
@click.argument(
    "directory",
    type=click.Path(dir_okay=True, exists=True, file_okay=False),
)
@pass_config
def background_cmd(
    cfg: Config,
    directory: str,
    workers: int | None,
    run_config: str | None,
    sleep: float,
    nice: int,
    no_email: bool = False,
) -> None:
    """Watch and run turnover jobs from directory"""
    from os import cpu_count
    from .background import SimpleQueue
    from .logger import logger

    if workers is None:
        workers = max((cpu_count() or 1) // 2, 1)
        logger.warning("using %d workers", workers)

    squeue = SimpleQueue(
        directory,
        workers=workers,
        wait=sleep,
        nice=nice,
        config=run_config or cfg.user_config,
        no_email=no_email,
    )
    squeue.run()



from __future__ import annotations

import operator
from collections.abc import Sequence
from dataclasses import dataclass
from dataclasses import replace
from typing import Any
from typing import Callable
from typing import Literal
from typing import NamedTuple

import numpy as np
import pandas as pd  # type: ignore
from sqlalchemy import and_
from sqlalchemy import func
from sqlalchemy import join
from sqlalchemy import or_
from sqlalchemy import Select
from sqlalchemy import select
from sqlalchemy import Subquery
from sqlalchemy import text
from sqlalchemy.engine import Connection
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.orm import Relationship
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.expression import BinaryExpression

from .model import PepProt
from .model import Peptide
from .model import Protein
from .utils import rehydrate_peptides
from .utils import round10

# from sqlalchemy.orm import load_only
# from sqlalchemy.orm import selectinload
# from sqlalchemy.orm import Session


# func.count()
# pylint: disable=not-callable
class WhereClauses(NamedTuple):
    pep_where: ColumnElement[bool] | None
    prot_where: ColumnElement[bool] | None


@dataclass
class FormFilter:
    peptideQualityFilter: float | None = None
    heavyMinCor: float | None = None
    isoPeaksNr: float | None = None
    isoPeaksMinRSQ: float | None = None
    #    heavyMinRSQ: float | None = None
    monoMinus1MinRatio: float | None = None
    # enrichmentFactor: float | None = None
    enrichmentMax: float | None = None
    enrichmentMin: float | None = None
    fdrMaximum: float | None = None
    nnlsResidual: float | None = None
    maxPeakArea: float | None = None
    log_sigma_lpf: float | None = None

    def where(self) -> WhereClauses:
        pep_filters = []
        prot_filters = []

        def add(q: Any) -> None:
            pep_filters.append(q)

        if self.peptideQualityFilter is not None:
            add(Peptide.peptideprophet_probability >= self.peptideQualityFilter)

        if self.fdrMaximum is not None:
            add(Peptide.fdr <= self.fdrMaximum)

        if self.heavyMinCor is not None:
            add(Peptide.heavyCor >= self.heavyMinCor)

        # if self.heavyMinRSQ is not None:
        #     add(Peptide.heavy_adj_r_squared >= self.heavyMinRSQ)

        if self.nnlsResidual is not None:
            add(Peptide.nnls_residual / Peptide.totalNNLSWeight <= self.nnlsResidual)

        if self.monoMinus1MinRatio is not None:
            add(Peptide.inv_ratio < 1.0 / self.monoMinus1MinRatio)

        if self.enrichmentMax is not None:
            add(Peptide.enrichment <= self.enrichmentMax)
        if self.enrichmentMin is not None:
            add(Peptide.enrichment >= self.enrichmentMin)

        if self.maxPeakArea is not None:
            add(Peptide.maxPeakArea >= self.maxPeakArea)

        if self.log_sigma_lpf is not None:
            prot_filters.append(
                func.log((Protein.lpf_std + 1e-4) / Protein.lpf_median)
                <= self.log_sigma_lpf,
            )

        if (
            self.isoPeaksNr is not None
            and self.isoPeaksMinRSQ is not None
            and self.isoPeaksMinRSQ != -1.0  # not set
        ):
            col = getattr(Peptide, f"iso_peaks_nr_{int(self.isoPeaksMinRSQ)}", None)
            if col is not None:
                add(col >= int(self.isoPeaksNr))

        return WhereClauses(
            and_(*pep_filters) if pep_filters else None,
            and_(*prot_filters) if prot_filters else None,
        )


def like(attr: InstrumentedAttribute, value: str) -> BinaryExpression[bool]:
    if "%" not in value:
        value = "%" + value + "%"
    return attr.like(value)


def doregex(attr: InstrumentedAttribute, value: str) -> BinaryExpression[bool]:
    return attr.op("REGEXP")(value)  # need pcre installed in sqllite


OPMAP: dict[str, Callable[..., Any]] = {
    ">": operator.gt,
    ">=": operator.ge,
    "=": operator.eq,
    "<": operator.lt,
    "<=": operator.le,
    "!=": operator.ne,
    "like": like,
    "regex": doregex,
}


def okcolumn(name: str, Orm: type[DeclarativeBase]) -> bool:
    attr = getattr(Orm, name)
    if attr is None or not hasattr(attr, "property"):
        return False
    if isinstance(attr.property, Relationship):
        return False
    return True


class RowQuery(NamedTuple):
    column: str
    op: str
    value: Any

    def to_sql(self, Orm: type[DeclarativeBase]) -> ColumnElement[bool]:
        attr = getattr(Orm, self.column)
        if attr is None or not hasattr(attr, "property"):
            raise ValueError(f"unknown column {self.column} for {Orm}")
        if isinstance(attr.property, Relationship):
            raise ValueError(f"column {self.column} is a relationship on {Orm}!")
        op = OPMAP.get(self.op)
        if op is None:
            raise ValueError(f"no operator like {self.op}")

        return op(attr, self.value)


@dataclass
class RowFilter:
    rows: list[RowQuery]
    method: Literal["and", "or"] = "or"

    def add(self, more: RowFilter) -> RowFilter:
        return replace(self, rows=self.rows + more.rows)

    def to_and_sql(self, Orm: type[DeclarativeBase]) -> ColumnElement[bool]:
        return and_(*[q.to_sql(Orm) for q in self.rows])

    def to_or_sql(self, Orm: type[DeclarativeBase]) -> ColumnElement[bool]:
        return or_(*[q.to_sql(Orm) for q in self.rows])

    def to_sql(self, Orm: type[DeclarativeBase]) -> ColumnElement[bool]:
        if self.method == "and":
            return self.to_and_sql(Orm)
        return self.to_or_sql(Orm)


@dataclass
class DTQuery:
    start: int = 0
    length: int = 10
    search: str | None = None
    regex: bool = False
    ascending: bool = True
    order_column: str | None = None
    draw: int = 0  # data tables

    # def query(self) -> re.Pattern | None:
    #     if not self.search:
    #         return None
    #     query = self.search if self.regex else f"^.*{re.escape(self.search)}.*$"
    #     return re.compile(query, re.I)

    # def like(self) -> str | None:
    #     if not self.search:
    #         return None
    #     return "%" + self.search + "%"


def search_protein_tosql(search: str, search_columns: list[str]) -> ColumnElement[bool]:
    return RowFilter(
        [RowQuery(coln, "like", search) for coln in search_columns],
    ).to_or_sql(Protein)


@dataclass
class ProteinQueryResult:
    result_df: pd.DataFrame
    total_proteins: int = -1
    total_peptides: int = -1
    total_filtered: int = -1


def np_agg(val: str | None, func: Callable[[np.ndarray], float]) -> float | None:
    if val is None:
        return None
    values = np.array([float(v) for v in val.split(",")])
    if len(values) == 0:
        return np.nan
    return func(values)


def std(val: str | None) -> float | None:
    return np_agg(val, np.std)


def median(val: str | None) -> float | None:
    return np_agg(val, np.median)


# sqlite is missing some aggregate functions... we "fake" them here.
FAKE_AGGREGATE_FUNCS = {"std": std, "median": median}


# see https://www.sqlite.org/lang_aggfunc.html
@dataclass
class Aggregate:
    column: str
    function: str  # aggregate function sum,avg,group_concat etc.
    label: str = ""
    args: Sequence[str] | str = ()

    def __post_init__(self) -> None:
        if self.label == "":
            self.label = self.column
        if isinstance(self.args, str):
            self.args = [self.args]

    def make_column(self, orm: type[DeclarativeBase]) -> InstrumentedAttribute[Any]:
        attr = getattr(orm, self.column)
        if attr is None or not hasattr(attr, "property"):
            raise ValueError(f"unknown column {self.column} for {orm}")
        if isinstance(attr.property, Relationship):
            raise ValueError(f"column {self.column} is a relationship on {orm}!")

        if self.function in FAKE_AGGREGATE_FUNCS:
            aggf = getattr(func, "group_concat")
            attr = aggf(attr, ",")
        else:
            aggf = getattr(func, self.function)
            attr = aggf(attr, *self.args)
        return attr.label(self.label)


class ProteinQuery:
    Join = join(PepProt, Peptide)
    # QID = select(PepProt.proteinid).select_from(Join)
    NPRO = select(func.count(PepProt.proteinid.distinct())).select_from(Join)
    COUNT_PROTEIN = select(func.count(Protein.proteinid))
    # NPEP = select(func.count(PepProt.peptideid.distinct())).select_from(Join)
    NPEP = select(func.count(Peptide.peptideid))

    def __init__(
        self,
        peptide_filter: FormFilter | None,
        protein_filter: RowFilter | None = None,
        *,
        query: DTQuery | None = None,
        search_columns: list[str] | None = None,
        protein_columns: list[str] | None = None,
        aggregates: Sequence[Aggregate] = (),
    ):
        self.protein_filter = protein_filter.to_sql(Protein) if protein_filter else None
        if peptide_filter is not None:
            self.peptide_filter, prot_where = peptide_filter.where()
            if prot_where is not None:
                self.protein_filter = (
                    and_(prot_where, self.protein_filter)
                    if self.protein_filter is not None
                    else prot_where
                )

        else:
            self.peptide_filter = None

        self.dtquery = query
        self.search_columns = search_columns
        self.aggregates = aggregates
        self.protein_columns = (
            [self.to_attr(col) for col in protein_columns]
            if protein_columns
            else [Protein]  # type: ignore
        )

        npro = self.NPRO
        # qid = self.QID
        npep = self.NPEP
        if self.peptide_filter is not None:
            npro = npro.where(self.peptide_filter)  # type: ignore
            # qid = qid.where(self.peptide_filter)  # type: ignore
            npep = npep.where(self.peptide_filter)  # type: ignore
        self.npro = npro
        # self.qid = qid
        self.npep = npep

    def make_agg(self, agg: Aggregate) -> InstrumentedAttribute[Any]:
        return agg.make_column(Peptide)

    def peptide_subquery(self, aggregates: Sequence[Aggregate]) -> Subquery:
        aggs = [self.make_agg(agg) for agg in aggregates]
        columns = [PepProt.proteinid, *aggs]
        q = select(*columns).select_from(self.Join)
        if self.peptide_filter is not None:
            q = q.where(self.peptide_filter)
        q = q.group_by(PepProt.proteinid)
        return q.subquery()

    def to_attr(self, col: str) -> InstrumentedAttribute[Any]:
        attr = getattr(Protein, col)
        if attr is None or not hasattr(attr, "property"):
            raise ValueError(f"unknown column {col} for protein table")
        if isinstance(attr.property, Relationship):
            raise ValueError(f"column {col} is a relationship on protein!")
        return attr

    def query(
        self,
        engine: Engine,
        want_all: bool = True,
    ) -> ProteinQueryResult:
        qid = self.peptide_subquery(self.aggregates)
        cols = qid.columns[1:]
        columns = self.protein_columns + list(cols)

        q: Select[Any] = select(*columns)
        q = q.join(qid, qid.c.proteinid == Protein.proteinid)
        # if self.peptide_filter is not None:
        #     q = q.where(Protein.proteinid.in_(self.qid))
        if self.protein_filter is not None:
            q = q.where(self.protein_filter)

        q, tq = self.process_query(q)

        tq = tq.join(qid, qid.c.proteinid == Protein.proteinid)
        total_proteins = total_peptides = total_filtered = -1
        with engine.begin() as conn:
            if want_all:
                total_proteins = conn.execute(self.npro).scalar() or 0
                total_peptides = conn.execute(self.npep).scalar() or 0
                total_filtered = conn.execute(tq).scalar() or 0
            df = pd.read_sql_query(q, con=conn)
            # sigh! sqllite does not have a std aggregate function
            # so we group_concat and the parse and the string again...
            for a in self.aggregates:
                if a.function in FAKE_AGGREGATE_FUNCS:
                    df[a.label] = df[a.label].apply(FAKE_AGGREGATE_FUNCS[a.function])

        return ProteinQueryResult(
            self.rehydrate(df),
            total_proteins,
            total_peptides,
            total_filtered,
        )

    def process_query(self, query: Select[Any]) -> tuple[Select[Any], Select[Any]]:
        q = self.dtquery
        if q:
            if q.search is not None and self.search_columns is not None:
                query = query.where(search_protein_tosql(q.search, self.search_columns))
            if q.order_column is not None:
                col = getattr(Protein, q.order_column, None)
                if col is not None:
                    query = query.order_by(col.asc() if q.ascending else col.desc())
                else:
                    # virtual
                    if q.order_column in {a.label for a in self.aggregates}:
                        asc = "ASC" if q.ascending else "DESC"
                        query = query.order_by(text(f"{q.order_column} {asc}"))

        filtered_total = self.COUNT_PROTEIN
        if query.whereclause is not None:
            filtered_total = filtered_total.where(query.whereclause)

        if q and q.length > 0:
            query = query.slice(q.start, q.start + q.length)
        return query, filtered_total

    # def process_model_query(self, query: Select[Any]) -> Select[Any]:
    #     if self.dtquery is None or self.search_columns is None:
    #         return query
    #     q = self.dtquery
    #     if q.search:
    #         query = query.where(search_protein_tosql(q.search, self.search_columns))
    #     if q.order_column is not None:
    #         col = getattr(Protein, q.order_column)
    #         query = query.order_by(col.asc() if q.ascending else col.desc())
    #     return query

    # def query_to_model(
    #     self,
    #     engine: Engine,
    #     peptide_columns: list[str] | None,
    # ) -> Sequence[Protein]:
    #     q = select(Protein)
    #     q = q.options(load_only(*self.protein_columns))

    #     options = selectinload(Protein.peptides)
    #     if peptide_columns:
    #         _cols = [getattr(Peptide, a) for a in peptide_columns]
    #         options = options.load_only(*_cols)
    #     q = q.options(options)

    #     if self.peptide_filter is not None:
    #         q = q.where(Protein.proteinid.in_(self.qid))
    #     if self.protein_filter is not None:
    #         q = q.where(self.protein_filter)
    #     q = self.process_model_query(q)
    #     with Session(bind=engine) as session:
    #         return session.execute(q).scalars().all()

    # def query_to_model_df(
    #     self,
    #     engine: Engine,
    #     peptide_cols: list[str] | None,
    #     peptides_to_column: Callable[[Protein, str], list[tuple[str, Any]]],
    # ) -> pd.DataFrame:
    #     models = self.query_to_model(engine, peptide_cols)

    #     df = pd.DataFrame(
    #         [
    #             {
    #                 **{k: getattr(p, k.key) for k in self.protein_columns or []},
    #                 **{"num_peptides": len(p.peptides)},
    #                 **dict(
    #                     t for k in peptide_cols or [] for t in peptides_to_column(p, k)
    #                 ),
    #             }
    #             for p in models
    #         ],
    #     )
    #     return df

    def rehydrate(self, df: pd.DataFrame) -> pd.DataFrame:
        return rehydrate_peptides(df)


class PeptideQuery:
    Join = join(PepProt, Protein)
    QID = select(PepProt.peptideid).select_from(Join)

    def __init__(
        self,
        peptide_filter: FormFilter | None = None,
        protein_filter: RowFilter | None = None,
        *,
        peptide_columns: list[str] | None = None,
        filtered_column: str | None = None,
    ):
        self.protein_filter = protein_filter.to_sql(Protein) if protein_filter else None

        qid = self.QID
        if self.protein_filter is not None:
            qid = qid.where(self.protein_filter)  # type: ignore

        self.qid = qid
        if peptide_filter is not None:
            clause = peptide_filter.where()
            self.peptide_filter = clause.pep_where
            # if clause.prot_where is not None:
            #     self.qid = self.qid.where(clause.prot_where)
        else:
            self.peptide_filter = None
        self.filtered_column = filtered_column
        self.peptide_columns = peptide_columns

    def query(
        self,
        engine: Engine,
    ) -> pd.DataFrame:
        cols = (
            [
                attr
                for col in self.peptide_columns
                for attr in [getattr(Peptide, col, None)]
                if attr is not None
            ]
            if self.peptide_columns
            else [Peptide]
        )

        if self.peptide_filter is not None and self.filtered_column is not None:
            cols.append(self.peptide_filter.label(self.filtered_column))

        q: Select[Any] = select(*cols)
        if self.protein_filter is not None:
            q = q.where(Peptide.peptideid.in_(self.qid))
        if self.peptide_filter is not None and self.filtered_column is None:
            q = q.where(self.peptide_filter)

        with engine.begin() as conn:
            q = self.process_query(q, conn)
            df = pd.read_sql_query(q, con=conn)
            return self.rehydrate(df)

    def process_query(self, q: Select[Any], conn: Connection) -> Select[Any]:
        return q

    def rehydrate(self, df: pd.DataFrame) -> pd.DataFrame:
        return rehydrate_peptides(df)


class SimplePeptideQuery:
    def __init__(
        self,
        peptide_filter: RowFilter | None = None,
        extra: FormFilter | None = None,
        *,
        peptide_columns: list[str] | None = None,
    ):
        self.peptide_filter = peptide_filter.to_sql(Peptide) if peptide_filter else None
        self.extra = extra
        self.peptide_columns = peptide_columns

    def query(
        self,
        engine: Engine,
    ) -> pd.DataFrame:
        q: Select[Any] = select(
            *(
                [
                    attr
                    for col in self.peptide_columns
                    for attr in [getattr(Peptide, col, None)]
                    if attr is not None
                ]
                if self.peptide_columns
                else [Peptide]
            ),
        )

        if self.peptide_filter is not None:
            q = q.where(self.peptide_filter)
        if self.extra is not None:
            wc = self.extra.where()
            if wc.pep_where is not None:
                q = q.where(wc.pep_where)
            # if wc.prot_where is not None:
            #     q = q.where(wc.prot_where)

        with engine.connect() as conn:
            df = pd.read_sql_query(q, con=conn)
            return self.rehydrate(df)

    def rehydrate(self, df: pd.DataFrame) -> pd.DataFrame:
        return rehydrate_peptides(df)


@dataclass
class Stats:
    max_nnls_residual: float = 1.0
    max_maxPeakArea: float = 0.0
    min_enrichment: float = 0.0
    max_enrichment: float = 1.0


def calc_stats(engine: Engine | None, round: bool = False) -> Stats:
    if engine is None:
        return Stats()
    q = select(
        func.max(Peptide.nnls_residual / Peptide.totalNNLSWeight).label(
            "max_nnls_residual",
        ),
        func.max(Peptide.maxPeakArea).label("max_maxPeakArea"),
        func.min(Peptide.enrichment).label("min_enrichment"),
        func.max(Peptide.enrichment).label("max_enrichment"),
    )
    with engine.connect() as conn:
        row = conn.execute(q).mappings().one()
    if round:
        d = {k: round10(v) for k, v in row.items()}
    else:
        d = dict(row)
    return Stats(**d)



from __future__ import annotations

from base64 import b64encode
from io import BytesIO

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy import stats

from .logger import logger
from .sns import despine
from .types.turnovertype import TurnoverDict
from .types.turnovertype import TurnoverRow
from .utils import ensure_pos
from .utils import isotopicDistribution
from .utils import PeptideSettings
from .utils import resize

dpi = 96
FIGSIZE: tuple[float, float] = (1056.0 / dpi, 768.0 / dpi)

EICS_CMAP = plt.get_cmap("viridis")

NEICS_CMAP = len(EICS_CMAP.colors)

EIC_COLUMNS = ["peptide", "modcol", "eics", "eics_shape", "monoFitParams"]


def peptide_legend(peptide: TurnoverRow, enrichment: float | None = None) -> str:
    if enrichment is None:
        enrichment = peptide.enrichment
    if enrichment is not None:
        ie = round(peptide.enrichment * 100, 1)
        ia = round(peptide.relativeIsotopeAbundance * 100, 1)
        return f"{peptide.peptide}\nLPF: {ia}%\nEnrichment: {ie}%"
    else:
        return f"{peptide.peptide},\nunable to estimate enrichment / LPF"


def despine_bottom(ax: Axes) -> None:
    despine(ax=ax, bottom=True, trim=True)


def despine_top(ax: Axes, offset: int = 5) -> None:
    despine(ax=ax, top=True, trim=True, offset=offset)


def plotEICS(
    peptide: TurnoverRow,
    ax: Axes | None = None,
    figsize: tuple[float, float] = FIGSIZE,
) -> Axes:
    from .fitenvelopes import normalIntensityFunction, FAIL_W_CURVE

    pep = peptide.peptide
    mod = peptide.modcol
    eics = peptide.eics
    monoFitParams = peptide.monoFitParams
    fail = len(monoFitParams) == 0  # old way
    if hasattr(peptide, "fail"):  # my way
        fail = fail | (getattr(peptide, "fail") & (1 << FAIL_W_CURVE))

    # maxInt = eics[0,:, 1].max()
    # ylim = (0, maxInt)
    if ax is None:
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    ax.set(xlabel="retention time", ylabel="Intensity", title=f"{pep} {mod}")

    m = eics[0]
    rt, im = m[:, 0], m[:, 1]
    ax.plot(rt, im, color="dodgerblue", lw=2)
    oeics = eics[2:]
    step = max(NEICS_CMAP // len(oeics), 1) if len(oeics) else 1

    for idx, e in enumerate(oeics):
        rt, i = e[:, 0], e[:, 1]
        icolor = min(idx * step, NEICS_CMAP - 1)
        ax.plot(
            rt,
            i,
            color=EICS_CMAP.colors[icolor],
            alpha=0.2,
            linestyle="dashed",
        )

    mono = eics[1]
    rt, im = mono[:, 0], mono[:, 1]
    ax.plot(rt, im, color="red", lw=2)
    d = dict(handlelength=0)
    if not fail:
        i = normalIntensityFunction(rt, *monoFitParams)
        ax.plot(rt, i, color="#8b2252", lw=2)  # violetred4
        ax.legend(labels=["Good"], labelcolor="darkgreen", **d)
    else:
        ax.legend(labels=["Bad"], labelcolor="salmon", **d)

    return ax


def intrange(n: int) -> np.ndarray:
    return np.arange(0, n, 1, dtype=np.int32)


LABELLED_COLUMNS = [
    "labelledEnvelopes",
    "labelledElementCount",
    "peptide",
    "enrichment",
    "relativeIsotopeAbundance",
    "theoreticalDist",
]


def plotLabelledEnvelope(
    peptide: TurnoverRow,
    settings: PeptideSettings,
    ax: Axes | None = None,
    scaled: bool = True,
    figsize: tuple[float, float] = FIGSIZE,
) -> Axes:
    from .fitenvelopes import get_enrichments

    if ax is None:
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    if len(peptide.labelledEnvelopes) > 0:
        e = peptide.labelledEnvelopes
        e = e if isinstance(e, np.ndarray) else np.array(e)
        x = intrange(len(e))
        enrichments = get_enrichments(peptide.peptide, settings)
        enrichments = resize(enrichments, len(e))
        if scaled:
            total = e.sum()
            e = e / total
        ax.bar(x, e, color="turquoise", edgecolor="black")
        ax.set_xticks(x)
        ax.set_xticklabels(
            [f"[{i}]{v:.2f}" for i, v in enumerate(enrichments)],
            rotation=-90,
        )
        ax.set(xlabel="enrichment")
        if scaled:
            ax.set(ylim=(0.0, 1.0), ylabel="fraction")

    else:
        ie = peptide.labelledElementCount
        x = intrange(ie + 1)
        ax.bar(x, np.repeat(0, ie + 1), color="turquoise")
        ax.set_xticks(x)
    ax.legend(labels=[peptide_legend(peptide)], handlelength=0)
    despine_bottom(ax)
    ax.xaxis.set_ticks_position("none")
    ax.set(title="labelledEnvelopes")
    return ax


FITTED_COLUMNS = [
    "peptide",
    "modcol",
    "isotopeEnvelopes",
    "maxIso",
    "relativeIsotopeAbundance",
    "enrichment",
    # "adjustedRsq",
]


def plotFittedEnvelopes(
    peptide: TurnoverRow,
    settings: PeptideSettings,
    ax: Axes | None = None,
    figsize: tuple[float, float] = FIGSIZE,
) -> Axes:
    pep = peptide.peptide
    mod = peptide.modcol
    isotopeEnvelopes = peptide.isotopeEnvelopes
    if len(isotopeEnvelopes) == 0:
        logger.warning("plotFittedEnvelopes[%s]: no envelope!", pep)
        return ax
    has_adj = hasattr(peptide, "adjustedRsq")
    if has_adj:
        adjustedRsq = resize(np.array(peptide.adjustedRsq), peptide.maxIso + 2)
    # theoreticalDist, naturalDist = norm(peptide, settings)

    isotopeEnvelopes = resize(isotopeEnvelopes, peptide.maxIso + 2)

    theoreticalDist = np.array(peptide.theoreticalDist)
    naturalDist = isotopeEnvelopes[1] * isotopicDistribution(peptide.peptide)

    naturalDist = resize(naturalDist, peptide.maxIso + 1)
    theoreticalDist = resize(theoreticalDist, peptide.maxIso + 1)

    x = intrange(len(isotopeEnvelopes)) - 1

    heavyDistribution = ensure_pos(isotopeEnvelopes[1:] - naturalDist)
    if ax is None:
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    xm1, x1 = x[:1], x[1:]  # skip negative heavy atom
    ax.set(title=f"{pep} {mod}")
    ax.bar(
        x1,
        isotopeEnvelopes[1:],
        color="turquoise",
        alpha=1.0,
    )
    ax.bar(x1, heavyDistribution, color="blue", width=0.5, alpha=0.3)
    if has_adj:
        ax.bar(x1, adjustedRsq[1:] * isotopeEnvelopes[1], color="tomato")

    ax.legend(labels=[peptide_legend(peptide)], handlelength=0, facecolor="none")
    ax.bar(xm1, isotopeEnvelopes[:1], color="red", alpha=0.3)
    ax.bar(x1, -theoreticalDist, color="#98F5FF", alpha=0.3)  # cadetblue1
    nd = resize(naturalDist, len(x) - 1)
    ax.vlines(x1, 0, nd, color="orange", linestyle="dashed")
    ax.scatter(x1, nd, color="orange", marker="o", s=20)

    xlabel = f"number of ${{}}^{{{settings.labelledIsotopeNumber}}}{settings.labelledElement}$ atoms"
    nfailed = np.isnan(isotopeEnvelopes[1:]).sum()
    if nfailed > 0:
        xlabel += f" [{nfailed} failed fits]"
    ax.set(xlabel=xlabel)

    ax.set_xticks(x)

    despine_bottom(ax)
    ax.xaxis.set_ticks_position("none")

    return ax


def plotIntensities(
    peptide: TurnoverRow,  # eics
    eici: list[int] | None = None,
    ax: Axes | None = None,
    figsize: tuple[float, float] = FIGSIZE,
) -> Axes:
    from scipy.linalg import lstsq

    intensity = peptide.eics[1, :, 1]
    im = np.c_[np.ones(len(intensity)), intensity]
    x = np.linspace(np.min(intensity), np.max(intensity), num=30)
    idx = set(range(peptide.eics.shape[0]))
    if eici is None:
        eici = list(idx)
    else:
        eici = list(sorted(set(eici) & idx))
    if ax is None:
        fig = Figure(figsize=figsize)
        ax = fig.subplots(1, 1)
    for i in eici:
        ints = peptide.eics[i, :, 1]
        fit, _residues, _rank, _s = lstsq(im, ints)

        y = fit[0] + fit[1] * x
        ax.scatter(intensity, ints, s=10)
        ax.plot(x, y)
    ax.set(
        title=f"{peptide.peptide} {peptide.modcol}",
        xlabel="mono intensity",
        ylabel="intensity",
    )
    return ax


def enrichment_plot(
    turn: pd.DataFrame,
    factor: float = 2.0,
    ax: Axes | None = None,
    figsize: tuple[float, float] | None = None,
    column: str = "enrichment",
) -> Axes:
    enrich = turn[column]
    enrich = enrich[enrich.notna()]
    em = enrich.mean()
    esd = enrich.std()
    if ax is None:
        fig = Figure(figsize=figsize if figsize else FIGSIZE)
        ax = fig.subplots(1, 1)

    x = np.linspace(enrich.min() - 3 * esd, enrich.max() + 3 * esd, 40)
    y = stats.norm(em, esd).pdf(x)
    ax.hist(
        enrich,
        bins=40,
        color="turquoise",
        alpha=0.4,
        density=True,
        edgecolor="#999",
    )
    ax.plot(x, y)
    ax.vlines([em - factor * esd, em + factor * esd], 0, [2.5, 2.5], color="red")
    despine_top(ax, offset=5)
    return ax


def nnls_values(
    df: pd.DataFrame,
    percent: float = 0.01,
) -> pd.Series:
    nnls_residual = df["nnls_residual"]
    if "totalNNLSWeight" in df.columns:
        nnls_residual = nnls_residual / df["totalNNLSWeight"]

    nnls_residual = nnls_residual[nnls_residual.notna()]
    if percent == 0.0:
        return nnls_residual

    bins, bounds = np.histogram(nnls_residual, bins=500)

    maxval = bins.max()
    imax: int = np.argmax(bins == maxval)  # type: ignore
    idx: int = min(len(bounds) - 1, imax + np.argmin(bins[imax:] / maxval > percent))  # type: ignore

    return nnls_residual[nnls_residual < bounds[idx]]


def nnls_plot(
    df: pd.DataFrame,
    ax: Axes | None = None,
    figsize: tuple[float, float] | None = None,
    percent: float = 0.01,
) -> Axes:
    nnls = nnls_values(df, percent=percent)

    if ax is None:
        fig = Figure(figsize=figsize if figsize else FIGSIZE)
        ax = fig.subplots(1, 1)
    # mn,mx = nnls_residual.min(), nnls_residual.max()
    ax.hist(nnls, bins="auto", color="turquoise")
    despine(ax, trim=True, offset=5)
    ax.set(ylabel="counts", xlabel="NNLS residual")
    return ax


def correlation_plot(
    df: pd.DataFrame,
    column: str,
    xlabel: str | None = None,
    ax: Axes | None = None,
    figsize: tuple[float, float] | None = None,
) -> Axes:
    if ax is None:
        fig = Figure(figsize=figsize if figsize else FIGSIZE)
        ax = fig.subplots(1, 1)
    # mn,mx = nnls_residual.min(), nnls_residual.max()
    col = df[column]
    ax.hist(col, bins="auto", color="turquoise")
    despine(ax, trim=True, offset=5)
    ax.set(ylabel="counts", xlabel=xlabel or column)
    return ax


ALL_COLUMNS = list(set(EIC_COLUMNS) | set(FITTED_COLUMNS) | set(LABELLED_COLUMNS))


def plot_all(
    peptide: TurnoverRow,
    settings: PeptideSettings,
    figsize: tuple[float, float] | None = None,
    hspace: float = 0.5,
    portrait: bool = True,
) -> tuple[Figure, list[Axes]]:
    if figsize is None:
        figsize = FIGSIZE[0], 3 * FIGSIZE[1]

    if not portrait:
        r, c = 1, 3
    else:
        r, c = 3, 1
    fig = Figure(figsize=figsize)
    axes = fig.subplots(r, c, gridspec_kw=dict(hspace=hspace))

    plotEICS(peptide, ax=axes[0])
    plotFittedEnvelopes(peptide, settings, ax=axes[1])
    plotLabelledEnvelope(peptide, settings, ax=axes[2])
    # plotIntensities(peptide, [1, 2, 3, 4, 5, 6], ax=ax4)
    return fig, axes


def rename(s: pd.Series, suffix: str = "_me") -> pd.Series:
    r = pd.Series(dtype=object)
    for key in [
        "isotopeEnvelopes",
        "relativeIsotopeAbundance",
        "monoFitParams",
        "eics",
        "labelledEnvelopes",
        "labelledElementCount",
        "enrichment",
        "heavyFitParams",
        "heavyCor",
    ]:
        r[key] = s[key + suffix]

    r.peptide = s.peptide
    r.modcol = s.modcol
    r.maxIso = int(s["maxIso" + suffix])
    return r


def plot_eics_sidebyside(
    b: TurnoverRow,
    suffix: str = "_me",
    figsize: tuple[float, float] = (15.0, 5.0),
) -> tuple[Figure, Axes, Axes]:
    fig = Figure(figsize=figsize)
    [ax1, ax2] = fig.subplots(1, 2)

    plotEICS(b, ax=ax1)
    plotEICS(rename(b, suffix), ax=ax2)
    return fig, ax1, ax2


def plot_fitted_sidebyside(
    b: TurnoverRow,
    settings: PeptideSettings,
    suffix: str = "_me",
    figsize: tuple[float, float] = (15, 5),
) -> tuple[Figure, Axes, Axes]:
    fig = Figure(figsize=figsize)
    [ax1, ax2] = fig.subplots(1, 2)

    plotFittedEnvelopes(b, settings, ax=ax1)
    plotFittedEnvelopes(rename(b, suffix), settings, ax=ax2)
    return fig, ax1, ax2


def plot_labelled_sidebyside(
    b: TurnoverRow,
    settings: PeptideSettings,
    suffix: str = "_me",
    figsize: tuple[float, float] = (15, 5),
) -> tuple[Figure, Axes, Axes]:
    fig = Figure(figsize=figsize)
    [ax1, ax2] = fig.subplots(1, 2)

    plotLabelledEnvelope(b, settings, ax=ax1)
    plotLabelledEnvelope(rename(b, suffix), settings, ax=ax2)
    return fig, ax1, ax2


def plot_fit_arrays(
    row: TurnoverRow,
    suffix: str = "_me",
    figsize: tuple[float, float] = (15, 12),
) -> tuple[Figure, list[list[Axes]]]:
    b: TurnoverDict = row  # type: ignore
    fig = Figure(figsize=figsize)
    axes = fig.subplots(
        5,
        2,
        gridspec_kw=dict(hspace=0.5),
    )
    [[ax1, ax2], [ax3, ax4], [ax5, ax6], [ax7, ax8], [ax9, ax10]] = axes

    def plot_a(name: str, ax: Axes) -> None:
        x = list(range(0, len(b[name])))  # type: ignore
        v = b[name + suffix]  # type: ignore
        if len(x) != len(v):
            ax.set(title=f"bad {name}")
            return
        dodge = np.nanmax(np.abs(v)) / 50
        ax.plot(x, b[name])  # type: ignore
        ax.plot(x, v + dodge)
        ax.set(title=f"{name}")

    def plot_a_col(name: str, ax: Axes, col: int) -> None:
        x = list(range(0, len(b[name][:, col])))  # type: ignore
        v = b[name + suffix][:, col]  # type: ignore
        if len(x) != len(v):
            ax.set(title=f"bad {name}")
            return
        dodge = np.nanmax(np.abs(v)) / 50
        ax.plot(x, b[name][:, col])  # type: ignore
        ax.plot(x, v + dodge)
        ax.set(title=f"{name}")

    # def plot_heavy_fit(ax: Axes) -> None:
    #     from .fitenvelopes import normalIntensityFunction2

    #     params = b["heavyFitParams"]
    #     if params is None or len(params) == 0:
    #         return

    #     hd = b["heavyDistribution"]

    #     x = np.array(list(range(len(hd))), dtype=np.float32)
    #     mu, sigma, k, r_sq, r_adj = params
    #     y: np.ndarray = normalIntensityFunction2(x, mu, sigma, k)  # type: ignore
    #     ax.plot(x, y, color="#5555ff")
    #     ax.text(x[0], y.max(), f"r_adj={r_adj:.2f}", fontsize=10)

    def ratio(name: str) -> np.ndarray:
        a = b[name]  # type: ignore
        m = b[name + suffix]  # type: ignore
        if len(a) != len(m):
            return a * 0
        both = (a == 0) & (m == 0)
        ret = m / np.where(a != 0, a, 1e-6)
        if both.sum() > 0:  # 0/0 is 1 !
            ret = np.where(both, 1.0, ret)
        return ret

    def plot_ratio(name: str, ax: Axes) -> None:
        val = b[name]  # type: ignore
        x = list(range(0, len(val)))
        ax.plot(x, ratio(name))
        ax.set(title=f"{name} ratio")

    plot_a_col("isotopeRegressions", ax1, 1)
    ax1.set(title="regression intensity")

    plot_a_col("isotopeRegressions", ax2, 0)
    ax2.set(title="adjusted R-squared")

    plot_a("isotopeEnvelopes", ax3)
    # plot_a("heavyDistribution", ax4)
    # plot_heavy_fit(ax4)
    plot_a("labelledEnvelopes", ax5)
    plot_a("monoFitParams", ax6)
    # plot_a("heavyFitParams", ax7)

    plot_ratio("monoFitParams", ax8)
    plot_ratio("isotopeEnvelopes", ax9)
    # plot_ratio("heavyDistribution", ax10)

    return fig, axes


# for use in jupyter notebooks....
def plot_all_sidebyside(
    b: TurnoverRow,
    settings: PeptideSettings,
) -> tuple[Figure, Figure, Figure, Figure]:
    f1, *_axes = plot_eics_sidebyside(b)
    f2, *_axes = plot_fitted_sidebyside(b, settings)
    f3, *_axes = plot_labelled_sidebyside(b, settings)
    f4, _axes = plot_fit_arrays(b)
    return f1, f2, f3, f4


def to_image_url(fig: Figure, format: str = "png") -> str:
    out = BytesIO()
    fig.savefig(out, format=format)
    data = f"data:image/{format.lower()};base64,".encode("ascii") + b64encode(
        out.getbuffer(),
    )
    return data.decode("ascii")


def plotLPF(
    df: pd.DataFrame,  # ['peptide', 'relativeIsotopeAbundance']
    ax: Axes | None = None,
    figsize: tuple[float, float] = (8.0, 1.5),
    sort: bool = True,
) -> Axes:
    if ax is None:
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    d = df.set_index("peptide")
    d = d["relativeIsotopeAbundance"]
    if sort:
        d = d.sort_values()
    # d.plot(ax=ax, style=".-", color="turquoise")
    d.plot(ax=ax, kind="bar", color="turquoise", edgecolor="black")
    ax.set(ylabel="LPF", ylim=(0, 1), xticks=[])
    despine(
        ax=ax,
        top=True,
        bottom=False,
        trim=True,
        offset=0,
    )
    legend = ax.get_legend()
    if legend is not None:
        legend.remove()

    ax.get_figure().tight_layout()
    return ax


def plotLogSigmaLPF(
    df: pd.DataFrame,
    ax: Axes | None = None,
    figsize: tuple[float, float] = (8.0, 4.0),
    remove_zero_std: bool = True,
) -> Axes:
    if ax is None:
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    key = df["lpf_median"] != 0.0
    if remove_zero_std:
        key = key & (df["lpf_std"] != 0.0)
    d = df[key]
    x = np.log((d["lpf_std"] + 1e-4) / (d["lpf_median"]))

    x.plot.hist(ax=ax, bins=100, color="turquoise", grid=False)
    despine(ax, trim=True, offset=5)
    ax.set(ylabel="counts", xlabel=r"$log(\sigma/LPF)$")
    return ax



from __future__ import annotations

import os
import queue
import select
import subprocess
import threading
from typing import Any
from typing import IO
from typing import Iterator


class Command:
    def __init__(
        self,
        argline: list[str],
        cwd: str | None = None,
        env: dict[str, str] | None = None,
        capture: bool = True,
        silent: bool = False,
    ):
        environ = dict(os.environ)
        if env:
            environ.update(env)
        kwargs: dict[str, Any] = {"cwd": cwd, "env": environ}
        if silent:
            kwargs["stdout"] = subprocess.DEVNULL
            kwargs["stderr"] = subprocess.DEVNULL
            capture = False
        elif capture:
            kwargs["stdout"] = subprocess.PIPE
            kwargs["stderr"] = subprocess.PIPE
        self.capture = capture
        self._cmd = subprocess.Popen(argline, **kwargs)

    def terminate(self) -> None:
        self._cmd.terminate()

    def wait(self) -> int:
        returncode = self._cmd.wait()
        return returncode

    @property
    def returncode(self) -> int:
        return self._cmd.returncode

    def __enter__(self) -> Command:
        return self

    def __exit__(self, exc_type, exc_value, tb) -> None:  # type: ignore
        self._cmd.wait()

    def __iter__(self) -> Iterator[str]:
        if not self.capture:
            raise RuntimeError("Not capturing")

        # Windows platforms do not have select() for files
        if os.name == "nt":
            q: queue.Queue[bytes] = queue.Queue()

            def reader(stream: IO[bytes]) -> None:
                while True:
                    line = stream.readline()
                    q.put(line)
                    if not line:
                        break

            t1 = threading.Thread(target=reader, args=(self._cmd.stdout,))
            t1.setDaemon(True)
            t2 = threading.Thread(target=reader, args=(self._cmd.stderr,))
            t2.setDaemon(True)
            t1.start()
            t2.start()
            outstanding = 2
            while outstanding:
                item = q.get()
                if not item:
                    outstanding -= 1
                else:
                    yield item.rstrip().decode("utf-8", "replace")

        # Otherwise we can go with select()
        else:
            streams = [self._cmd.stdout, self._cmd.stderr]
            while streams:
                for fps in select.select(streams, [], streams):
                    for stream in fps:
                        line = stream.readline()
                        if not line:
                            if stream in streams:
                                streams.remove(stream)
                            break
                        yield line.rstrip().decode("utf-8", "replace")

    def safe_iter(self) -> Iterator[str]:
        with self:
            yield from self

    @property
    def output(self) -> Iterator[str]:
        return self.safe_iter()

    @property
    def pid(self) -> int:
        return self._cmd.pid



from __future__ import annotations

import logging
import mmap
import re
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Iterator
from typing import NamedTuple
from typing import overload
from typing import TypedDict

import numpy as np
import pandas as pd

from .fitenvelopes import fitEnvelopes
from .jobs import TurnoverJob
from .logger import logger
from .parallel_utils import Task
from .types.pepxmltypes import PepXMLRunRowRT
from .utils import BaseResourceFile
from .utils import getsize
from .utils import human
from .utils import IO
from .utils import MzMLResourceFile
from .utils import MzMLResourceFileLocal
from .utils import PeptideSettings
from .utils import PepXMLResourceFile
from .utils import ProtXMLResourceFile
from .utils import ResourceFiles
from .utils import ResultsResourceFile


class D(TypedDict):
    retention_time_sec: np.ndarray
    scanindex: np.ndarray
    mzmax: np.ndarray
    mzmin: np.ndarray
    imax: np.ndarray


SPECTRUM_QUERY = re.compile(b"<spectrum ")


def scan_mzml_spectra(mzml: Path) -> Iterator[int]:
    """Very fast inspection of .mzML spectra queries"""
    with mzml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in SPECTRUM_QUERY.finditer(mm):
                yield m.start(0)


class Search(NamedTuple):
    mzmin: float
    mzmax: float
    rtmin: float
    rtmax: float

    def isoverlap(self, other: Search) -> bool:
        dmz = min(self.mzmax, other.mzmax) - max(self.mzmin, other.mzmin)
        if dmz < 0:
            return False
        drt = min(self.rtmax, other.rtmax) - max(self.rtmin, other.rtmin)
        if drt < 0:
            return False
        return True


class ByteOrderError(ValueError):
    pass


def mzi_max_intensity(mzi: np.ndarray, mzmin: float, mzmax: float) -> float:
    mz, i = mzi[0], mzi[1]
    q = (mz <= mzmax) & (mz >= mzmin)
    return i[q].max() if q.any() else 0.0


def zeroi() -> np.ndarray:
    return np.array([], dtype=np.float32).reshape(MZI.SHAPE)


COLS = ("mzmin", "mzmax", "rtmin", "rtmax")


class MZI:
    SHAPE = (2, -1)

    def __init__(
        self,
        mzmlfile: MzMLResourceFile,
    ):
        self.name = mzmlfile.name
        meta, mapname = mzmlfile.cache_mzml(), mzmlfile.cache_memmap()

        self._mzi = np.memmap(mapname, dtype=np.float32, mode="r")

        if self._mzi[0] != MAGIC_NO:
            raise ByteOrderError(
                f'file "{mapname}" can\'t be read (written with different byteorder)',
            )
        # columns: retention_time_sec', 'scanindex', 'mzmax', 'mzmin', 'imax'
        self.df = IO(meta).read_df()

        scanstart = np.concatenate(  # pylint: disable=unexpected-keyword-arg
            ([1], self.df["scanindex"].to_numpy(dtype=np.int64)[:-1]),
            dtype=np.int64,
        )
        self.df["scanstart"] = scanstart
        mzmax = self.df["mzmax"].max()
        mzmin = self.df["mzmin"].min()
        rtmin = self.df["retention_time_sec"].min()
        rtmax = self.df["retention_time_sec"].max()
        self.search = Search(mzmin, mzmax, rtmin, rtmax)

        self.dino = None
        # dino = mzmlfile.cache_dinosaur()
        # if dino.exists():
        #     self.dino = IO(dino).read_df(columns=["mz", "rtStart", "rtEnd"])
        # else:
        #     self.dino = None

    def __repr__(self) -> str:
        return (
            f"MZI({self.name}[{len(self.df)}],"
            f" mz=[{self.search.mzmin:.2f},{self.search.mzmax:.2f}],"
            f" rt=[{self.search.rtmin:.2f},{self.search.rtmax:.2f}])"
        )

    def mzi_column(self, df: pd.DataFrame | None = None) -> pd.Series:
        if df is None:
            df = self.df
        return df[["scanstart", "scanindex"]].apply(self.mzi_apply, axis=1)

    def mzi(self, scanstart: int, scanindex: int) -> np.ndarray:
        return self._mzi[scanstart:scanindex].reshape(self.SHAPE)

    def mzi_apply(self, row: pd.Series) -> np.ndarray:
        return self.mzi(row["scanstart"], row["scanindex"])

    @classmethod
    def from_path(cls, path: str) -> MZI:
        return cls(MzMLResourceFile(path))

    @classmethod
    def from_file(cls, mzmlfile: str, directory: str | None = None) -> MZI:
        return cls(MzMLResourceFileLocal(mzmlfile, directory))

    @classmethod
    def can_read_mzi(cls, mzi: str | Path) -> bool:
        _mzi = np.memmap(mzi, dtype=np.float32, mode="r")
        return _mzi[0] == MAGIC_NO

    def getEIC(
        self,
        mzrange: np.ndarray,  # float[maxIso+2,2]
        rtmin: float,
        rtmax: float,
        *,
        minrt: int = 0,
    ) -> np.ndarray | None:  # float[maxIso+2,rt,2]
        # ensure within boundary
        s = self.search
        if not (
            rtmin >= s.rtmin
            and rtmax <= s.rtmax
            and np.any(mzrange[:, 0] <= s.mzmax)
            and np.any(mzrange[:, 1] >= s.mzmin)
        ):
            return None
        rt = self.df["retention_time_sec"]
        df = self.df[(rt >= rtmin) & (rt <= rtmax)]
        if len(df) < minrt:
            return None

        mzi = self.mzi_column(df)
        rt = df["retention_time_sec"]
        out = np.zeros((len(mzrange), len(rt), 2), dtype=np.float32)
        for idx, (mzmin, mzmax) in enumerate(mzrange):
            outofrange = (df["mzmax"] < mzmin) | (df["mzmin"] > mzmax)
            iz = mzi[outofrange].apply(lambda i: 0.0)
            inz = mzi[~outofrange].apply(mzi_max_intensity, args=(mzmin, mzmax))
            # stack columns and reset index to rt
            imax = pd.concat(
                [idf for idf in [iz, inz] if len(idf) > 0],
                axis=0,
            ).reindex(index=rt.index)
            # 2 column array [rt, intensity]
            rti = pd.concat([rt, imax], axis=1)
            out[idx, :, :] = rti.to_numpy(dtype=np.float32)

        if out[:, :, 1].max() <= 0.0:
            return None
        return out

    def _getEIC(
        self,
        mzmin: float,
        mzmax: float,
        rtmin: float,
        rtmax: float,
    ) -> np.ndarray | None:
        if not self.search.isoverlap(Search(mzmin, mzmax, rtmin, rtmax)):
            return None

        rt = self.df["retention_time_sec"]
        t = (
            (rt >= rtmin)
            & (rt <= rtmax)
            & (self.df["mzmin"] <= mzmax)
            & (self.df["mzmax"] >= mzmin)
        )
        if t.sum() == 0:
            return None

        filtered_df = self.df[t]

        def toimax(row: pd.Series) -> float:
            return mzi_max_intensity(self.mzi_apply(row), mzmin, mzmax)

        imax = filtered_df[["scanstart", "scanindex"]].apply(
            toimax,
            axis=1,
        )

        if imax.max() <= 0.0:
            return None

        rti = pd.concat([filtered_df["retention_time_sec"], imax], axis=1)
        # array will be in column major 'F' order so
        # that transpose will be ok
        return rti.to_numpy(dtype=np.float32)

    def eics_from_df(
        self,
        df: pd.DataFrame,
        columns: list[str] | None = None,
        *,
        zero_length: bool = False,
    ) -> pd.Series:
        cols = COLS
        if columns:
            df = df.rename(
                columns=dict(zip(columns, cols)),
            )

        def findeic(row: pd.Series) -> np.ndarray | float:
            ret = self._getEIC(row[cols[0]], row[cols[1]], row[cols[2]], row[cols[3]])
            if ret is None:
                return zeroi() if zero_length else np.nan
            return ret.T

        return df[list(cols)].apply(findeic, axis=1)

    @classmethod
    def dehydrate_eics(cls, s: pd.Series) -> pd.Series:
        return s.apply(lambda a: a.flatten() if isinstance(a, np.ndarray) else a)

    @classmethod
    def rehydrate_eics(cls, s: pd.Series) -> pd.Series:
        return s.apply(
            lambda a: a.reshape(cls.SHAPE) if isinstance(a, np.ndarray) else a,
        )

    def __len__(self) -> int:
        return len(self.df)

    @overload
    def __getitem__(self, idx: int) -> np.ndarray:
        ...

    @overload
    def __getitem__(self, idx: slice) -> pd.Series:
        ...

    @overload
    def __getitem__(self, idx: list[int]) -> pd.Series:
        ...

    def __getitem__(self, idx: int | list[int] | slice) -> np.ndarray | pd.Series:
        df = self.df[["scanstart", "scanindex"]].iloc[idx]
        if isinstance(idx, int):
            scanstart, scanindex = df
            return self.mzi(scanstart, scanindex)

        return df.apply(self.mzi_apply, axis=1)  # .to_numpy()


# number must survive roundtrip....
# struct.unpack("f", struct.pack("f", MAGIC_NO))[0] == MAGIC_NO
MAGIC_NO = -0.1234000027179718
# MAGIC_BYTES = b'$\xb9\xfc\xbd'
MAGIC_BYTES = np.array([MAGIC_NO], dtype=np.float32).tobytes()


def mzml_create(
    mzml: MzMLResourceFile,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> int:
    """Create cache files for a mzML file"""
    from .broken_api import PyMzMLReader
    from .logger import log_iterator

    mzreader = PyMzMLReader(
        mzml.original,
        build_index_from_scratch=False,
    )
    total = mzreader.get_spectrum_count()

    d: D = dict(
        retention_time_sec=np.zeros(total, dtype=np.float32),
        scanindex=np.zeros(total, dtype=np.int64),
        mzmax=np.zeros(total, dtype=np.float32),
        mzmin=np.zeros(total, dtype=np.float32),
        imax=np.zeros(total, dtype=np.float32),
    )
    n = 0
    scanindex = 1  # first part is a magic number
    mname = mzml.cache_memmap()

    with mname.open("wb") as fp:
        it = log_iterator(
            mzreader,
            total=total,
            desc=mzml.original.name,
            level=level,
            number_of_bg_processes=number_of_bg_processes,
        )

        fp.write(MAGIC_BYTES)
        # loop over <spectrumList count="nnnn"><spectrum>....</spectrum>... </spectrumList>
        for spectrum in it:
            # <cvParam cvRef="MS" accession="MS:1000511" name="ms level" value="1"/>
            if spectrum.ms_level == 1:
                # <cvParam cvRef="MS" accession="MS:1000016" name="scan start time" value="0.001483586933"
                #       unitCvRef="UO" unitAccession="UO:0000031" unitName="minute"/>
                rt = spectrum.scan_time_in_minutes() * 60.0
                mz, i = spectrum.mz, spectrum.i
                mzi = np.row_stack((mz, i))
                # mz first, intensity second
                data = mzi.flatten().astype(np.float32)  # .tobytes()
                scanindex += len(data)
                # we will use these to search for suitable rt,mz from pep.xml files
                d["retention_time_sec"][n] = rt
                d["mzmax"][n] = mz.max()
                d["mzmin"][n] = mz.min()

                d["imax"][n] = i.max()
                d["scanindex"][n] = scanindex
                n += 1
                fp.write(data.tobytes())

    df = pd.DataFrame({k: v[:n] for k, v in d.items()})  # type: ignore
    df.sort_values(by="retention_time_sec", inplace=True)
    df.reset_index(drop=True, inplace=True)

    mzml_out(mzml, df)

    from .dinosaur.dinosaur import mzml_dinosaur

    mzml_dinosaur(mzml)

    return level


def mzml_out(
    mzml: MzMLResourceFile,
    df: pd.DataFrame,
) -> None:
    out = mzml.cache_mzml()
    IO(out, df).save_df()

    if logger.isEnabledFor(logging.INFO):
        mname = mzml.cache_memmap()
        mem = df.memory_usage(deep=True).sum()
        g = getsize
        msg = (
            f"written: {mzml.original.name} -> {out.name}: memory={human(mem)} ondisk={human(g(out))},"
            f" mzimap={human(g(mname))} original={human(g(mzml.original))}"
        )
        logger.info(msg)


def dfinfo(df: pd.DataFrame, mzml: MzMLResourceFile, out: Path) -> None:
    if logger.isEnabledFor(logging.INFO):
        mem = df.memory_usage(deep=True).sum()
        g = getsize
        ret = f"{out.name}: memory={human(mem)} ondisk={human(g(out))} original={human(g(mzml.original))}"
        logger.info(ret)


def dehydrate_eics(df: pd.DataFrame) -> pd.DataFrame:
    df["eics_shape"] = df["eics"].apply(lambda a: a.shape)
    df["eics"] = df["eics"].apply(lambda e: e.flatten())
    if "isotopeRegressions" in df.columns:
        df["isotopeRegressions"] = df["isotopeRegressions"].apply(lambda e: e.flatten())
    return df


def turnover_run(
    job: TurnoverJob,
    jobrun: ResultsResourceFile,
    *,
    workers: int = 4,
    save_subset: bool = True,
    nspectra: int | None = None,
) -> None:
    from .pepxml import process_pep_prot, dehydrate_pepxml
    from .utils import human, rmfiles
    from .exts import EICS
    from .parallel_utils import parallel_tasks

    start = datetime.now()
    pepxml_df = process_pep_prot(job)

    tasks = create_mz_tasks(pepxml_df, job, nspectra=nspectra)
    if logger.isEnabledFor(logging.INFO):
        taskmem = sum(t.mem() for t in tasks)
        mem = pepxml_df.memory_usage(deep=True).sum()
        info = f"pepxml memory={human(mem)}, tasks[{len(tasks)}/{workers}] total memory=[{human(taskmem)}]"
        logger.info(info)
    # save
    if save_subset:
        pepxml_df = pd.concat(
            [pepxml_df.loc[task.pepxml_df.index] for task in tasks],
            axis="index",
        )
        pepxml_df.index.name = "pepxml_index"
        pepxml_df = pepxml_df.reset_index()
    pepxml_df_filename = jobrun.cache_pepxml()
    pepxml_df = dehydrate_pepxml(pepxml_df)
    IO(pepxml_df_filename, pepxml_df).save_df()
    pepxml_df = None
    # get mzml tasks
    envelopes: list[Path] = []

    ntotal = len(tasks)
    try:
        for idx, (df, mzml) in enumerate(parallel_tasks(tasks, workers=workers), 1):
            logger.info(
                "EIC[%s] task done[%d/%d]: %s[%d]",
                jobrun.original.name,
                idx,
                ntotal,
                mzml.original.name,
                len(df),
            )
            if len(df) == 0:
                continue

            eicsfile = jobrun.cache_file(f"-{idx}{EICS}")
            df = dehydrate_eics(df)
            IO(eicsfile, df).save_df()
            envelopes.append(eicsfile)

            dfinfo(df, mzml, eicsfile)

        save_result(
            jobrun.cache_result(),
            pepxml_df_filename,
            job.to_resource_files().protxml,
            envelopes,
        )
        logger.info("turnover finished! %s after %s", job.jobid, datetime.now() - start)
    finally:
        envelopes.append(pepxml_df_filename)
        rmfiles(envelopes)


def save_result(
    result_name: Path,
    pepdf: Path,
    protdf: ProtXMLResourceFile,
    envelopes: list[Path],
) -> None:
    from .sqla.model import save_to_file
    from .protxml import getprotxml

    logger.info("consolidating eics")
    eic = consolidate_eic(envelopes)
    eic = eic.set_index("peptide_index")

    me = IO(pepdf).read_df()
    prot = getprotxml(protdf)
    n = len(me)
    if "pepxml_index" in me.columns:  # in save subset
        me = me.set_index("pepxml_index")
    me = me.join(eic, how="inner", rsuffix="_env")
    me = me.reset_index(drop=True)
    if "peptide_env" in me.columns:
        assert (me["peptide"] == me["peptide_env"]).all()
        me.drop(columns=["peptide_env"], inplace=True)
    # create explicit index for parquet
    me.index.name = "result_index"
    me.reset_index(inplace=True)  # add result_index to columns
    # TODO join on what?
    # me = me.join(prot, how="inner")
    # name, *_ = result_name.name.split(".")
    logger.info(
        "writing results: %s[%d] original length=%d",
        result_name.name,
        len(me),
        n,
    )
    # IO(result_name, me).save_df()
    me = dedup_peptides(me)
    save_to_file(result_name, me, prot)


# def dedup_peptides1(pep: pd.DataFrame) -> pd.DataFrame:
#     return (
#         pep.groupby(by=["peptide", "modcol"])
#         .apply(lambda x: x.sort_values(by="heavyCor", ascending=False).head(1))
#         .reset_index(drop=True)
#     )


def dedup_peptides(pep: pd.DataFrame) -> pd.DataFrame:
    """Remove duplicate peptide/modcol rows, taking smallest nnls_residual."""
    return (
        pep.sort_values(by="nnls_residual", ascending=True)
        .groupby(by=["peptide", "modcol"])
        .head(1)
        .reset_index()
    )


def consolidate_eic(eicfiles: list[Path]) -> pd.DataFrame:
    df = pd.concat(
        [IO(eic).read_df() for eic in eicfiles],
        axis="index",
        ignore_index=True,
    )
    return df


def verify_run(
    job: TurnoverJob,
) -> bool:
    from .types.checking import check_pepxml_columns

    files = job.to_resource_files()
    if not files.protxml.cache_protxml_ok():
        msg = "cache is out of date" if files.protxml.exists else "no prot XML file"
        logger.info("%s: %s", files.protxml.original.name, msg)
        ret = False

    ret = True
    df = None
    for pepxml in files.pepxmls:
        if not pepxml.cache_pepxml_ok():
            msg = "cache is out of date" if pepxml.exists else "no pepxml file"
            logger.info("%s: %s", pepxml.original.name, msg)
            ret = False
        else:
            fname = pepxml.cache_pepxml()
            df = IO(fname).read_df()
            missing = check_pepxml_columns(df)
            if missing:
                ret = False
                for m in missing:
                    logger.info("%s: %s", fname, m)
            n = len(df)

    total = 0
    failed = [mzml for mzml in files.mzmlfiles if not mzml.cache_mzml_ok()]
    if failed:
        for mzml in failed:
            msg = "cache is out of date" if mzml.exists else "no mzML file"
            logger.info("%s: %s", mzml.original.name, msg)
            ret = False

    else:
        if job.match_runNames:
            mzfile_to_run = job.get_mzfile_to_run()
            for mzml in files.mzmlfiles:
                if df is not None:
                    hits = (df["run"] == mzfile_to_run[mzml.name]).sum()
                else:
                    hits = 0
                total += hits
                logger.info("%s: %6d/%d: %s", mzml.original.name, hits, n, mzml.name)
        if df is not None and logger.isEnabledFor(logging.INFO):
            logger.info(
                "%s: %6d/%d",
                ", ".join(pepxml.original.name for pepxml in files.pepxmls),
                total,
                n,
            )

    return ret


class MzMLTask(Task):
    def task_run(self) -> pd.DataFrame:
        mzmlc = MZI(self.mzml)
        done = mzml_calc_mzml_envelopes(
            self.pepxml_df,
            mzmlc,
            self.settings,
            level=self.level,
            number_of_bg_processes=self.number_of_bg_processes,
        )

        outdf = pd.DataFrame(done)
        for col in ["relativeIsotopeAbundance", "enrichment", "heavyCor"]:
            if col in outdf.columns:
                outdf[col] = outdf[col].astype(np.float32)
        return outdf


def create_mz_tasks(
    pepxml_df: pd.DataFrame,
    job: TurnoverJob,
    nspectra: int | None = None,  # spectra chunk sizes for tasks
) -> list[MzMLTask]:
    tasks = []
    mz2run = job.get_mzfile_to_run()
    files = job.to_resource_files()
    todo: list[tuple[pd.DataFrame, MzMLResourceFile]] = []

    for mzml in files.mzmlfiles:
        if job.match_runNames:
            df = pepxml_df[pepxml_df["run"] == mz2run[mzml.name]]
        else:
            df = pepxml_df

        if len(df) == 0:
            continue
        # Don't need "peptide" really... just for sanity check
        df = df[
            [
                "retention_time_sec",
                "peptide",
                "maxIso",
                "mzranges",
                "isLabelledRun",
            ]
        ]
        if nspectra is not None and nspectra > 0:
            chunk_num = max(len(df) // nspectra, 1)
            if chunk_num > 1:
                dfs = [(sdf, mzml) for sdf in np.array_split(df, chunk_num)]
            else:
                dfs = [(df, mzml)]
        else:
            dfs = [(df, mzml)]
        todo.extend(dfs)

    ntotal = len(todo)
    tasks = [
        MzMLTask(cdf, mzml, job.settings, level, ntotal)
        for level, (cdf, mzml) in enumerate(todo, start=1)
    ]

    return tasks


def calc_rtranges(
    rundf: pd.DataFrame,
    mzmlc: MZI,
    settings: PeptideSettings,
    rt_min: int,
) -> pd.DataFrame:
    rt = mzmlc.df["retention_time_sec"]
    if mzmlc.dino is not None:
        rundf = rundf.join(findrt_dino(mzmlc.dino, rundf, rt, settings))
        rrt = rundf.rtmax - rundf.rtmin
        logger.info("dinosaur: rt range %s-%s secs", str(rrt.min()), str(rrt.max()))
    else:
        rundf = rundf.join(findrt(rundf, rt, settings))
    s = mzmlc.search
    q = (
        (rundf["rtmin"] <= s.rtmax)
        & (rundf["rtmax"] >= s.rtmin)
        & (rundf["rt_count"] >= rt_min)
        & rundf["mzranges"].apply(
            lambda mzr: np.any(mzr[:, 0] <= s.mzmax) and np.any(mzr[:, 1] >= s.mzmin),
        )
    )

    return rundf[q]


def findrt(
    rundf: pd.DataFrame,
    rt: pd.Series,
    settings: PeptideSettings,
) -> pd.DataFrame:
    use_in_sample = settings.retentionTimeCorrection == "UseInSample"
    rttol = settings.rtTolerance

    def rtrange(row: pd.Series) -> pd.Series:
        rts = row.retention_time_sec  # from pep.xml file
        if use_in_sample and row.isLabelledRun:
            rtmin = rts - rttol / 2
            rtmax = rts + rttol / 2

        else:
            rtmin = rts - rttol
            rtmax = rts + rttol

        rt_count = ((rt >= rtmin) & (rt <= rtmax)).sum()

        return pd.Series(
            dict(rtmin=rtmin, rtmax=rtmax, rt_count=float(rt_count)),
            dtype=np.float32,
        )

    return rundf[["retention_time_sec", "isLabelledRun"]].apply(rtrange, axis=1)


def findrt_dino(
    dino_df: pd.DataFrame,
    rundf: pd.DataFrame,
    rt: pd.Series,
    settings: PeptideSettings,
) -> pd.DataFrame:
    use_in_sample = settings.retentionTimeCorrection == "UseInSample"

    dinomz = dino_df["mz"]
    rttol = settings.rtTolerance

    def findrtrange(row: pd.Series) -> pd.Series:
        total = 0
        rtmin = 1e10
        rtmax = -1.0

        for mzin, mzmax in row.mzranges:
            q = (dinomz >= mzin) & (dinomz <= mzmax)
            n = q.sum()
            if n > 0:
                df = dino_df[q]
                rtmin = min(rtmin, df.rtStart.min())
                rtmax = max(rtmax, df.rtEnd.max())
            total += n
            if not total:
                if use_in_sample and row.isLabelledRun:
                    rtmin = row.retention_time_sec - rttol / 2
                    rtmax = row.retention_time_sec + rttol / 2
                else:
                    rtmin = row.retention_time_sec - rttol
                    rtmax = row.retention_time_sec + rttol
            rt_count = ((rt >= rtmin) & (rt <= rtmax)).sum()
        return pd.Series(
            dict(rtmin=rtmin, rtmax=rtmax, rt_count=float(rt_count)),
            dtype=np.float32,
        )

    return rundf[["mzranges", "retention_time_sec", "isLabelledRun"]].apply(
        findrtrange,
        axis=1,
    )


def mzml_calc_mzml_envelopes(
    rundf: pd.DataFrame,
    mzmlc: MZI,
    settings: PeptideSettings,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> list[dict[str, Any]]:
    from .logger import log_iterator
    from .config import MIN_RT

    n = len(rundf)
    rundf = calc_rtranges(
        rundf,
        mzmlc,
        settings,
        rt_min=MIN_RT,
    )
    logger.info(
        "%s[%d/%d]: removed out of range rt,mz %d/%d",
        mzmlc.name,
        level,
        number_of_bg_processes,
        len(rundf),
        n,
    )
    total = len(rundf)
    it = log_iterator(
        rundf.itertuples(index=True),
        total=total,
        desc=mzmlc.name,
        level=level,
        number_of_bg_processes=number_of_bg_processes,
    )
    row: PepXMLRunRowRT
    done = []
    # nhits = 0
    for row in it:
        idx = row[0]
        d = mzml_calc_pep_envelopes(mzmlc, row, settings)
        if d is None:
            continue
        d["peptide_index"] = idx
        d["peptide"] = row.peptide  # don't need this just sanity check
        # d['runName'] = mzmlc.mzmlfile.runName

        done.append(d)
    logger.info(
        "%s[%d/%d]: found %d/%d envelopes",
        mzmlc.name,
        level,
        number_of_bg_processes,
        len(done),
        total,
    )
    return done


def mzml_calc_pep_envelopes(
    mzmlc: MZI,
    row: PepXMLRunRowRT,
    settings: PeptideSettings,
) -> dict[str, Any] | None:
    from .config import MIN_RT

    rtmin, rtmax = row.rtmin, row.rtmax

    # mono = row.mzranges[1].reshape(-1, 2)
    # eic = mzmlc.getEIC(mono, rtmin, rtmax, minrt=MIN_RT, step=XCMS_STEP)
    # if eic is not None:
    #     res = fitMonoEnvelope(row.peptide, eic[0])
    #     if res is not None:
    #         rtmin, rtmax = res

    start = datetime.now()
    eic = mzmlc.getEIC(row.mzranges, rtmin, rtmax, minrt=MIN_RT)
    if eic is None:
        return None
    # if all mzrange,rt are zero
    # FIXME: more severe culling....
    if np.all(eic[:, :, 1] == 0.0):
        return None
    end = datetime.now()

    logger.debug("getEIC: %s %s %d", row.peptide, end - start, len(eic))

    maxIso = row.maxIso
    assert len(eic) == maxIso + 2

    start = end
    e = fitEnvelopes(row, eic, settings)
    if e is None:
        return None
    logger.debug("fitEnvelopes: %s %s", row.peptide, datetime.now() - start)
    d = dict(eics=eic)
    d.update(asdict(e))
    return d


def turnover_prepare(
    files: ResourceFiles,
    force: bool = False,
    workers: int = 1,
) -> None:
    """prepare pepXML and mzmlFiles"""
    from functools import partial
    from .pepxml import pepxml_create
    from .protxml import protxml_create
    from .parallel_utils import parallel_result

    if len(files.mzmlfiles) == 0:
        logger.warning("no mzML files!")
        return

    files.ensure_directories()

    procs: list[
        tuple[
            Callable[[Any, int, int], int],
            MzMLResourceFile | PepXMLResourceFile | ProtXMLResourceFile,
        ]
    ] = []

    for pepxml in files.pepxmls:
        if force or not pepxml.cache_pepxml_ok():
            procs.append((pepxml_create, pepxml))

        else:
            logger.info(
                'skipping creation of "%s" cache: %s',
                pepxml.name,
                pepxml.cache_pepxml().name,
            )
    if force or not files.protxml.cache_protxml_ok():
        procs.append((protxml_create, files.protxml))

    else:
        logger.info(
            'skipping creation of "%s" cache: %s',
            files.protxml.name,
            files.protxml.cache_protxml().name,
        )
    todo = (
        [t for t in files.mzmlfiles if not t.cache_mzml_ok()]
        if not force
        else files.mzmlfiles
    )
    skipped = set(files.mzmlfiles) - set(todo)
    if skipped:
        for t in skipped:
            logger.info(
                'skipping creation of "%s" cache: %s',
                t.name,
                t.cache_mzml().name,
            )
    target: MzMLResourceFile | PepXMLResourceFile | ProtXMLResourceFile
    for target in todo:
        procs.append((mzml_create, target))

    if not procs:
        return
    cleanups: dict[int, BaseResourceFile] = {}
    exe: list[Callable[[], int]] = []
    for idx, (func, target) in enumerate(procs, start=1):
        exe.append(partial(func, target, idx, len(procs)))
        cleanups[idx] = target

    ntotal = len(exe)
    try:
        for ridx in parallel_result(exe, workers=workers):
            logger.info("turnover_prepare task done: [%d/%d]", ridx, ntotal)
            if ridx in cleanups:
                del cleanups[ridx]

    except KeyboardInterrupt:
        for c in cleanups.values():
            c.cleanup()
        raise



from __future__ import annotations

from dataclasses import dataclass
from typing import cast

import numpy as np
import pandas as pd


def and_filters(filters: list[np.ndarray]) -> np.ndarray:
    if not filters:
        raise ValueError("no filters")
    if len(filters) == 1:
        return filters[0]
    return np.bitwise_and.reduce(filters)


class BaseFilter:
    @classmethod
    def columns(cls) -> list[str]:
        return []

    def filter(self, df: pd.DataFrame, *, copy: bool = False) -> pd.DataFrame:
        b = self.booleans(df)  # pylint: disable=assignment-from-none
        if b is None:
            return df
        return df[b].copy() if copy else df[b]

    def booleans(self, df: pd.DataFrame) -> pd.Series | None:
        return None

    def and_filters(self, filters: list[np.ndarray]) -> np.ndarray:
        return and_filters(filters)


@dataclass
class PepXMLFilter(BaseFilter):
    noX: bool = True
    minProbabilityCutoff: float | None = 0.8

    @classmethod
    def columns(cls) -> list[str]:
        return ["peptide", "peptideprophet_probability", "probability"]

    def booleans(self, df: pd.DataFrame) -> pd.Series | None:
        filters: list[np.ndarray] = []

        def add(b: bool) -> None:
            filters.append(cast(np.ndarray, b))

        if self.noX:
            add(~df["peptide"].str.contains("X"))

        if self.minProbabilityCutoff:
            if "peptideprophet_probability" in df:
                add(df["peptideprophet_probability"] >= self.minProbabilityCutoff)
            elif "probability" in df:
                add(df["probability"] >= self.minProbabilityCutoff)

        return pd.Series(self.and_filters(filters))



from __future__ import annotations

import logging
import mmap
import re
from collections import Counter
from os.path import commonprefix
from pathlib import Path
from typing import Callable
from typing import Iterator

import numpy as np
import pandas as pd

from .jobs import TurnoverJob
from .logger import logger
from .utils import Apply
from .utils import human
from .utils import IO
from .utils import PeptideSettings
from .utils import PepXMLResourceFile


def chop_spectrum(spectrum: str) -> str:
    ret = ".".join(spectrum.split(".")[:-3])
    return ret


def bchop_spectrum(spectrum: bytes) -> bytes:
    ret = b".".join(spectrum.split(b".")[:-3])
    return ret


SPECTRUM = re.compile(b'spectrum="([^"]+)"')


def count_spectra(pepxml: Path) -> dict[str, int]:
    """Very fast inspection of pep.xml spectra names"""
    cnt: dict[bytes, int] = Counter()
    with pepxml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in SPECTRUM.finditer(mm):
                spectrum = bchop_spectrum(m.group(1))
                cnt[spectrum] += 1

    return {k.decode("ascii"): v for k, v in cnt.items()}


SPECTRUM_QUERY = re.compile(b"<spectrum_query ")


def scan_spectra(pepxml: Path) -> Iterator[int]:
    """Very fast inspection of pep.xml spectra queries"""
    with pepxml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in SPECTRUM_QUERY.finditer(mm):
                yield m.start(0)


PEPTIDE_PROPHET_QUERY = re.compile(b"<peptideprophet_summary ")


def scan_pp_probability(pepxml: Path) -> bool:
    with pepxml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for _ in PEPTIDE_PROPHET_QUERY.finditer(mm):
                return True
    return False


def fractionExtractorRegexFinder(runNames: list[str]) -> re.Pattern | None:
    if len(runNames) <= 1:
        return None
    prefix = commonprefix(runNames)
    n = len(prefix)
    rev = [f[n:][::-1] for f in runNames]
    postfix = commonprefix(rev)[::-1]
    if n + len(postfix) >= len(min(runNames, key=len)):
        return None
    return re.compile(f"^{re.escape(prefix)}(.*){re.escape(postfix)}$")


def decoy_prefix_fn(prefix: str = "DECOY_") -> Callable[[list[str]], bool]:
    prefix = prefix.lower()
    return lambda proteins: all(p.lower().startswith(prefix) for p in proteins)


def decoy_postfix_fn(postfix: str = "_DECOY") -> Callable[[list[str]], bool]:
    postfix = postfix.lower()
    return lambda proteins: all(p.lower().endswith(postfix) for p in proteins)


def init_agg(
    df: pd.DataFrame,
    decoy_prefix: str = "DECOY_",
    decoy_postfix: str | None = None,
) -> pd.DataFrame:
    from .utils import calculate_fdrs

    decoyfn = (
        decoy_postfix_fn(decoy_postfix)
        if decoy_postfix is not None
        else decoy_prefix_fn(decoy_prefix)
    )
    df["modcol"] = df["modifications"].apply(Apply.modcol)
    df["is_decoy"] = df["proteins"].apply(decoyfn)
    if "fdr" not in df.columns:
        for prob in ["peptideprophet_probability", "probability"]:
            if prob in df.columns:
                df["fdr"] = calculate_fdrs(df, score=prob)
                break
    # Before or after grouping!

    # add H+
    df["observed_mz"] = (
        df["precursor_neutral_mass"] / df["assumed_charge"] + 1.00727646677
    )

    df["mz"] = df["calc_neutral_pep_mass"] / df["assumed_charge"] + 1.00727646677

    return df


def aggregateSearchHits(
    df: pd.DataFrame,
    *,
    settings: PeptideSettings,
    decoy_prefix: str = "DECOY_",
    decoy_postfix: str | None = None,
    runNames: set[str] | None = None,
) -> pd.DataFrame:
    from .config import GROUP_RT

    mztol = settings.mzTolerance
    rttol = settings.mzTolerance
    use_simple_median = settings.retentionTimeCorrection != "UseInSample"

    df = init_agg(
        df,
        decoy_prefix=decoy_prefix,
        decoy_postfix=decoy_postfix,
    )
    if runNames is not None:
        df["isLabelledRun"] = df["run"].apply(lambda r: r in runNames)
        keys = ["run"]
    else:
        df["isLabelledRun"] = False
        keys = []

    df["agg_count"] = 1

    keys.extend(["peptide", "_massint", "assumed_charge", "modcol"])
    df["_massint"] = df["calc_neutral_pep_mass"].apply(lambda m: round(m / mztol))
    remove = ["_massint"]

    if GROUP_RT:
        df["_rtint"] = df["retention_time_sec"].apply(lambda rt: round(rt / rttol))
        keys.append("_rtint")
        remove.append("_rtint")

    # cols = ['spectrum', 'spectrumNativeID', 'precursor_neutral_mass',
    #    'assumed_charge', 'retention_time_sec', 'start_scan', 'end_scan',
    #    'index', 'proteins', 'protein_descr', 'peptide_next_aa',
    #    'peptide_prev_aa', 'num_tol_term', 'xcorr', 'deltacn', 'deltacnstar',
    #    'spscore', 'sprank', 'expect', 'modifications', 'hit_rank', 'peptide',
    #    'num_tot_proteins', 'num_matched_ions', 'tot_num_ions',
    #    'num_missed_cleavages', 'calc_neutral_pep_mass', 'massdiff',
    #    'num_matched_peptides', 'modified_peptide', 'fval', 'ntt', 'nmc',
    #    'massd', 'isomassd', 'peptideprophet_probability',
    #    'peptideprophet_ntt_prob', 'is_rejected']

    def unique_seq(some_list: list[list[str]]) -> list[str]:
        return list({s for lv in some_list for s in lv})

    def join(some_list: list[str]) -> str:
        return ":".join(sorted(set(some_list)))

    # Lei's simple 'use local' rt correction
    def labelled_rt(rts: pd.Series) -> float:
        lr = df.loc[rts.index].isLabelledRun
        if lr.any():
            return rts[lr].median()
        return rts.median()

    agg = dict(
        calc_neutral_pep_mass="median",
        precursor_neutral_mass="median",
        mz="median",
        observed_mz="median",
        retention_time_sec="median" if use_simple_median else labelled_rt,
        num_missed_cleavages="max",
        peptideprophet_probability="max",
        interprophet_probability="max",
        is_decoy="all",
        agg_count="sum",
        # is_rejected="any",
        proteins=unique_seq,
        protein_descr=unique_seq,
        modifications="first",
        xcorr="max",  # COMET searchEngineScore
        ionscore="max",  # MASCOT searchEngineScore
        run=join,
        isLabelledRun="any",
        searchEngineScore="max",
        fdr="max",
    )

    agg_protxml = dict(
        probability="max",
        group_number="nunique",
        percent_coverage="max",
        confidence="max",
        protein_description=unique_seq,
        proteins=unique_seq,
        ngroups="max",
    )
    agg.update(agg_protxml)
    # only values that exist
    agg = {k: v for k, v in agg.items() if k in df.columns and k not in keys}
    cols = keys + list(agg)
    g = df[cols].groupby(by=keys, as_index=False)

    ret = g.aggregate(agg)
    # if not GROUP_RT:
    #     ret = ret.drop_duplicates(subset=keys + ["retention_time_sec"], keep="first")
    ret["prStr"] = ret["proteins"].apply(lambda proteins: ", ".join(sorted(proteins)))
    ret["description"] = ret["protein_descr"].apply(
        lambda protein_descr: ", ".join(sorted(protein_descr)),
    )
    ret["prGroup"] = ret.groupby("prStr").ngroup()
    # remove.append("prStr")
    ret = ret.drop(columns=remove)
    return ret


def pepxml_raw(
    pepxml: PepXMLResourceFile,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> pd.DataFrame:
    # from .broken_api import PepXMLDataFrame
    from .pepxml_reader import pepxml_dataframe

    logger.info("reading: %s", pepxml.original.name)
    df = pepxml_dataframe(
        pepxml.original,
        level=level,
        number_of_bg_processes=number_of_bg_processes,
    )

    logger.info("done: %s", pepxml.original.name)
    return df


def compute_settings(
    df: pd.DataFrame,
    settings: PeptideSettings,
    abundanceCutoff: float,
) -> pd.DataFrame:
    from .utils import (
        isotopicDistribution,
        peptideFormula,
        NAMES,
        labelledAtomicAbundances2,
    )
    from .utils import eic_mzranges
    from .types.pepxmltypes import PepXMLRunRow

    element = NAMES == settings.labelledElement
    assert element.sum() == 1

    def element_count(peptide: str) -> int:
        return peptideFormula(peptide)[element][0]

    # labelled abundances at maximum Enrichment level
    abundances = labelledAtomicAbundances2(settings)

    # maxIso length of ndist[E] to 0.01
    def max_iso(peptide: str) -> int:
        return len(isotopicDistribution(peptide, abundances, abundanceCutoff)) - 1

    def mzranges(pep: PepXMLRunRow) -> np.ndarray:
        return eic_mzranges(pep, settings)

    logger.info(
        "compute_settings: adding labelledElementCount, maxIso, mzranges",
    )
    df["labelledElementCount"] = df["peptide"].apply(element_count)
    df["maxIso"] = df["peptide"].apply(max_iso)
    # ndarray[maxIso+2,2] of mzmin,mzmax
    df["mzranges"] = df[["maxIso", "assumed_charge", "mz", "observed_mz"]].apply(
        mzranges,
        axis=1,
    )
    logger.info("compute_settings: done")
    return df


def getpepxml(
    pepxml: PepXMLResourceFile,
) -> pd.DataFrame:
    if pepxml.cache_pepxml_ok():
        data = pepxml.cache_pepxml()
        logger.info('getpepxml: reading: "%s"', data.name)
        df = IO(data).read_df()
        logger.info('getpepxml: finished reading: "%s" [%d]', data.name, len(df))

    else:
        if not pepxml.exists:
            raise RuntimeError(f"can't find file: {pepxml.original}")
        logger.info('getpepxml: creating cache "%s"', pepxml.original.name)
        pepxml_create(pepxml)
        assert pepxml.cache_pepxml_ok()
        df = IO(pepxml.cache_pepxml()).read_df()
    return df


def filter_pepxml(df: pd.DataFrame, job: TurnoverJob) -> pd.DataFrame:
    from .config import ABUNDANCE_CUTOFF
    from .filters import PepXMLFilter

    if job.match_runNames:
        df = df[df["run"].isin(job.runNames)].copy()
        logger.info('filter_pepxml: run names: "%s" [%d]', job.runNames, len(df))

    filter = PepXMLFilter(minProbabilityCutoff=job.settings.minProbabilityCutoff)
    df = filter.filter(df, copy=True)

    if "agg_count" not in df.columns:
        logger.info("aggregating: %d", len(df))
        df = aggregateSearchHits(
            df,
            settings=job.settings,
            runNames=job.runNames,
        )
        logger.info("aggregating done: %d", len(df))

    df = compute_settings(df, job.settings, abundanceCutoff=ABUNDANCE_CUTOFF)
    # df = df.iloc[0:100]  # DEBUG
    return df


# def join_pep_prot(
#     pepxml_df: pd.DataFrame,
#     protxml_df: pd.DataFrame,
#     settings: PeptideSettings,
# ) -> pd.DataFrame:
#     # TODO: join peptide and protxml
#     return pepxml_df


def process_pep_prot(job: TurnoverJob) -> pd.DataFrame:
    files = job.to_resource_files()
    ret = []
    for pepxml in files.pepxmls:
        pepxml_df = getpepxml(pepxml)

        pepxml_df = filter_pepxml(pepxml_df, job)
        ret.append(pepxml_df)
    pepxml_df = pd.concat(ret, axis=0, ignore_index=True)

    return pepxml_df


def pepxml_out(
    df: pd.DataFrame,
    pepxml: PepXMLResourceFile,
) -> None:
    from .utils import IO

    data = pepxml.cache_pepxml()

    if logger.isEnabledFor(logging.INFO):
        if "agg_count" in df.columns:
            logger.info(  # pylint: disable=logging-fstring-interpolation
                f"writing file: {data.name}: total={len(df)} multiple={(df.agg_count > 1).sum()}",
            )
        else:
            logger.info("writing file: %s: total=%s", data.name, len(df))

    df = dehydrate_pepxml(df)
    IO(data, df).save_df()

    if logger.isEnabledFor(logging.INFO):
        mem = df.memory_usage(deep=True)
        size = data.stat().st_size
        osize = pepxml.original.stat().st_size
        runs = ",".join(sorted(df.run.unique()))
        logger.info(  # pylint: disable=logging-fstring-interpolation
            f"memory={human(mem.sum())} disk={human(size)} original={human(osize)}: runs={runs}",
        )


def pepxml_create(
    pepxml: PepXMLResourceFile,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> int:
    df = pepxml_raw(
        pepxml,
        level=level,
        number_of_bg_processes=number_of_bg_processes,
    )
    pepxml_out(df, pepxml)
    return level


def dehydrate_pepxml(df: pd.DataFrame) -> pd.DataFrame:
    if "mzranges" in df.columns:
        df["mzranges"] = df["mzranges"].apply(lambda x: x.flatten())
    return df


def pepok(target: PepXMLResourceFile, force: bool) -> bool:
    if not force and target.cache_pepxml_ok():
        return True
    return False



from __future__ import annotations

import os
import signal
import subprocess
import sys
import time
from datetime import datetime
from os import walk
from os.path import join
from pathlib import Path
from shutil import which
from typing import Any
from typing import Iterator

import tomli as tomllib
import tomli_w

from .logger import logger

try:
    from psutil import pid_exists

    psutil_ok = True
except ImportError:
    psutil_ok = False

    # for windows we need psutil.pid_exists(pid)
    def pid_exists(pid: int) -> bool:
        """Check whether pid exists in the current process table Unix Only."""
        if pid == 0:
            # According to "man 2 kill" PID 0 has a special meaning:
            # it refers to <<every process in the process group of the
            # calling process>> so we don't want to go any further.
            # If we get here it means this UNIX platform *does* have
            # a process with id 0.
            return True
        try:
            os.kill(pid, 0)
        except ProcessLookupError:
            return False
        except PermissionError:
            # EPERM clearly means there's a process to deny access to
            return True
        # According to "man 2 kill" possible error values are
        # (EINVAL, EPERM, ESRCH)
        else:
            return True


# from https://stackoverflow.com/questions/18499497/how-to-process-sigterm-signal-gracefully
class GracefulKiller:
    kill_now: bool = False

    def __init__(self) -> None:
        self.kill_now: bool = False
        self.old: Any = None

    def __enter__(self) -> None:
        self.old = signal.signal(signal.SIGTERM, self.exit_gracefully)

    def __exit__(self, *args: Any) -> None:
        if self.old is not None:
            signal.signal(signal.SIGTERM, self.old)
            self.old = None
        if self.kill_now:
            raise ContinueException()

    def exit_gracefully(self, *args: Any) -> None:
        self.kill_now = True


class ContinueException(Exception):
    pass


class Handler:
    def __init__(self, processing: bool = True):
        self.processing = processing

    def __call__(self, signum: int, frame: Any) -> None:
        if self.processing:
            return
        raise ContinueException()


PID_FILENAME = "turnover.pid"
PROCESSING = "turnover.pid.is_running"


class Processing:
    """Is the background worker processing a job ATM?"""

    def __init__(self, jobsdir: Path):
        self._processing = jobsdir.joinpath(PROCESSING)

    def is_processing(self) -> bool:
        return self._processing.exists()

    def set_processing(self, start: bool) -> None:
        if start:
            self._processing.touch()
        else:
            try:
                self._processing.unlink(missing_ok=True)
            except OSError:
                pass


def signal_me(kill:bool=False) -> signal.Signals:
    if sys.platform == "win32":
        if kill:
            return signal.CTRL_BREAK_EVENT
        return signal.SIGBREAK
    return signal.SIGCONT


class SimpleQueueClient:
    """Used by web client to interact with background Queue.

    Currently only just sends a signal to wake up.
    """

    def __init__(self, jobdir: Path):
        self.jobdir = jobdir
        self.pidfile = jobdir.joinpath(PID_FILENAME)

    def terminate(self) -> bool:
        pid = self.get_pid()
        if pid is None:
            return False
        try:
            os.kill(pid, signal.SIGINT)
        except ProcessLookupError:
            return False
        except PermissionError:
            return False
        return True

    def signal(self) -> bool:
        pid = self.get_pid()
        if pid is None:
            return False
        try:
            os.kill(pid, signal_me(True))
        except ProcessLookupError:
            return False
        except PermissionError:
            return False
        return True

    def is_running(self) -> bool:
        pid = self.get_pid()
        if pid is None:
            return False
        return self._pid_exists(pid)

    def get_pid(self) -> int | None:
        if self.pidfile.exists():
            with self.pidfile.open("r") as fp:
                try:
                    return int(fp.read())
                except (TypeError, OSError):
                    return None
        return None

    def _pid_exists(self, pid: int) -> bool:
        return pid_exists(pid)


class SimpleQueue:
    INTERROR = 22

    def __init__(
        self,
        jobdir: str,
        wait: float = 60,
        workers: int = 4,
        nice: int = 0,
        config: str | None = None,
        no_email: bool = False,
    ):
        self.jobdir = jobdir
        self.wait = wait
        self.workers = workers
        self.nice = nice
        self.no_email = no_email
        self.nice_cmd = which("nice")
        self.handler = Handler()
        if config is not None:
            if not os.path.exists(config):
                logger.warning(
                    "configuration file %s doesn't exist! ignoring...",
                    config,
                )
                config = None
        self.config = config
        self.processing = Processing(Path(jobdir))
        signal.signal(signal_me(), self.handler)

    def command(self, *args: str) -> list[str]:
        nice = []
        if self.nice_cmd and self.nice:
            nice = [self.nice_cmd, "-n", str(self.nice)]

        config = [f"--config={self.config}"] if self.config is not None else []

        no_email = ["--no-email"] if self.no_email else []

        return [
            *nice,
            sys.executable,
            "-m",
            "protein_turnover",
            *config,
            "run",
            f"--interrupt-as-error={self.INTERROR}",
            f"--workers={self.workers}",
            *no_email,
            *args,
        ]

    def pidfile(self, tomlfile: Path) -> Path:
        return tomlfile.parent.joinpath(tomlfile.name + ".pid")

    def rmlog(self, tomlfile: Path) -> None:
        lf = tomlfile.parent.joinpath(tomlfile.stem + ".log")
        lf.unlink(missing_ok=True)

    def runjobs(self, it: Iterator[Path]) -> None:
        for tomlfile in it:
            self.rmlog(tomlfile)

            with subprocess.Popen(
                self.command(str(tomlfile)),
                shell=False,
                text=True,
            ) as proc:
                try:
                    logger.info("%s: running pid=%d", tomlfile, proc.pid)
                    # let website know that this job is running....
                    # see SimpleQueueClient
                    pid = self.pidfile(tomlfile)
                    with pid.open("w") as fp:
                        fp.write(str(proc.pid))
                    self.processing.set_processing(True)
                    try:
                        _, errs = proc.communicate()
                        if errs:
                            logger.error("error from %s: %s", tomlfile, errs)
                        ret = proc.wait()  # TODO what happens when we're terminated?
                    finally:  # can catch KeyboardInterrupt
                        pid.unlink(missing_ok=True)
                        self.processing.set_processing(False)
                    if ret < 0:
                        if -ret in {signal.SIGTERM, signal.SIGKILL}:
                            logger.warning("%s: killed...", tomlfile)
                    status = (
                        "finished"
                        if ret == 0
                        else ("killed" if ret < 0 or ret == self.INTERROR else "failed")
                    )
                    logger.info("%s: status=%s", tomlfile, status)
                    try:
                        self.update_status(tomlfile, status)
                    except Exception as e:
                        logger.error("can't update status! %s: %s", tomlfile, e)
                except KeyboardInterrupt:
                    logger.info("sending signal to child process")
                    proc.send_signal(
                        signal.SIGINT
                        if sys.platform != "win32"
                        else signal.CTRL_C_EVENT
                    )
                    raise

    def update_status(self, tomlfile: Path, status: str) -> None:
        c = self.read_toml(tomlfile)
        if c is None:
            return
        c["status"] = status
        with tomlfile.open("wb") as fp:
            tomli_w.dump(c, fp)

    def read_toml(self, tomlfile: Path) -> dict[str, Any] | None:
        try:
            with tomlfile.open("rb") as fp:
                return tomllib.load(fp)
        except Exception as e:
            logger.error("can't open %s: %s", tomlfile, e)
        return None

    def status(self, tomlfile: Path) -> str:
        c = self.read_toml(tomlfile)
        if c is None:
            return "failed"
        return str(c.get("status", "stopped"))

    def search(self, directory: str, wait: float = 60.0) -> Iterator[Path]:
        nloop = 0
        mtime = None
        todo: list[tuple[Path, datetime]] = []
        every = max(60 * int(60 / wait), 1) if wait else 1  # each hour
        while True:
            if (nloop % every) == 0:
                logger.info("searching for jobs in %s", directory)
            nloop += 1
            for d, _, files in walk(directory):
                for f in files:
                    if f.endswith(".toml"):
                        tomlfile = Path(join(d, f))
                        # already running
                        if self.pidfile(tomlfile).exists():
                            continue
                        mod = datetime.fromtimestamp(tomlfile.stat().st_mtime)
                        if mtime is None or mod > mtime:
                            try:
                                if self.status(tomlfile) == "pending":
                                    todo.append((tomlfile, mod))
                            except (
                                Exception  # pylint: disable=broad-exception-caught
                            ) as e:
                                logger.error("%s: %s", tomlfile, e)

            if todo:
                n = len(todo)
                logger.info("found %d job%s", n, "" if n == 1 else "s")
                todo = sorted(todo, key=lambda t: t[1])  # oldest first
                for tomlfile, m in todo:
                    if tomlfile.exists():
                        mtime = m
                        yield tomlfile
                todo = []
            try:
                self.handler.processing = False
                time.sleep(wait)
            except ContinueException:
                # kill -CONT $(cat {pidfile})
                logger.info("awakened by signal....")
            finally:
                self.handler.processing = True

    def process(self, jobdir: str, wait: float = 60.0) -> None:
        self.runjobs(self.search(jobdir, wait))

    def run(self) -> None:
        pidfile = Path(self.jobdir).joinpath(PID_FILENAME)
        try:
            with pidfile.open("wt", encoding="utf-8") as fp:
                pid = os.getpid()
                fp.write(str(pid))
            w = "with 👍" if psutil_ok else "without 👎"
            print(f"protein_turnover queue running ({w} psutil) as pid={pid}")
            self.process(self.jobdir, self.wait)
        finally:
            pidfile.unlink(missing_ok=True)



from __future__ import annotations

from collections import defaultdict
from math import fabs
from typing import Any
from typing import Callable
from typing import Iterator
from typing import Literal
from typing import NamedTuple
from typing import TypedDict

import pandas as pd
from lxml import etree


# umod = "http://www.unimod.org/xmlns/schema/unimod_2"
URL = "http://www.unimod.org/xml/unimod.xml"

ROWS = [
    "unimod_id",
    "name",
    "description",
    "mono_delta",
    "average_delta",
    "composition",
    "approved",
    "specificities",
]


class Composition(NamedTuple):
    symbol: str
    number: int


class DeltaD(TypedDict):
    mono_mass: float
    avge_mass: float
    composition: list[Composition]


class Unimod(NamedTuple):
    unimod_id: int
    name: str
    approved: bool
    description: str
    specificities: list[Specificity]
    mono_delta: float
    average_delta: float
    composition: list[Composition]


AZ = {chr(s) for s in range(ord("A"), ord("Z") + 1)}


class Specificity(NamedTuple):
    site: str  # Literal["Anywhere", "N-term", "C-term", *AZ]
    position: Literal[
        "Anywhere",
        "Any N-term",
        "Any C-term",
        "Protein N-term",
        "Protein C-term",
    ]
    classification: str
    isCommon: bool
    comment: str | None = None

    def matches(
        self,
        sequence: str,
        modPos: int,
        isProteinTerm: str | None = None,
    ) -> bool:
        def matchSite() -> bool:
            site = self.site
            if site == "Anywhere":
                return True
            if site == "Any N-term":
                return modPos == 1
            if site == "Any C-term":
                return modPos == len(sequence)
            return sequence[modPos - 1] == site

        position = self.position
        if position == "Anywhere":
            return matchSite()
        if position == "Any N-term":
            return modPos == 1 and matchSite()
        if position == "Any C-term":
            return modPos == len(sequence) and matchSite()
        if position == "Protein N-term":
            return isProteinTerm == "N" and matchSite()
        if position == "Protein C-term":
            return isProteinTerm == "C" and matchSite()

        return False


def fetch_unimod_from_web(filename: str = "unimod.xml") -> str:
    import requests

    resp = requests.get(URL, timeout=10)
    unimod_bytes = resp.content
    print("writing", filename)
    with open(filename, "wb") as fp:
        fp.write(unimod_bytes)
    return filename


class UnimodLookup:
    def __init__(
        self,
        unimods: list[Unimod] | dict[float, list[Unimod]],
        knownMods: Any = None,  # from pepxml file TODO
        massTolerance: float = 2e-5,
    ):
        self.knownMods = knownMods
        self.massTolerance = massTolerance
        if not isinstance(unimods, dict):
            unimodd = defaultdict(list)
            for u in unimods:
                unimodd[u.mono_delta].append(u)

            self.unimods: dict[float, list[Unimod]] = {k: v for k, v in unimodd.items()}
        else:
            self.unimods = unimods

    def create(self, knownMods: Any) -> UnimodLookup:
        return UnimodLookup(self.unimods, knownMods, self.massTolerance)

    def matchModification(
        self,
        sequence: str,
        position: int,
        monoDelta: float,
        isProteinTerm: Literal["N", "C"] | None = None,  # N,C or None
    ) -> list[tuple[Unimod, bool]]:
        """Returns a list or uniprot rows."""
        matchingMassMods = list(
            filter(lambda t: fabs(t - monoDelta) < self.massTolerance, self.unimods),
        )
        if not matchingMassMods:
            return []

        mod_delta_best = sorted(matchingMassMods, key=lambda d: fabs(d - monoDelta))[0]
        ret = []
        for mod in self.unimods[mod_delta_best]:
            matches = [
                s.isCommon
                for s in mod.specificities
                if s.matches(sequence, position, isProteinTerm)
            ]
            if matches:
                ret.append((mod, any(matches)))
        # most common first (this is different from Josh)
        return sorted(ret, key=lambda t: 0 if t[1] else 1)


def nodes(n: etree.Element) -> str:
    return etree.tostring(n).decode("utf-8")


def parse_unimod(root: etree.Element) -> Iterator[Unimod]:
    def specd(spec: etree.Element) -> Specificity:
        notes = " ".join(
            n.text for n in spec.xpath(".//umod:misc_notes", namespaces=root.nsmap)
        )
        d = dict(spec.attrib)

        site = d["site"]
        assert site in {"Anywhere", "N-term", "C-term"} or site in AZ, (
            site,
            nodes(spec),
        )

        pos = d["position"]
        assert pos in {
            "Anywhere",
            "Any N-term",
            "Any C-term",
            "Protein N-term",
            "Protein C-term",
        }, (
            pos,
            nodes(spec),
        )

        d["isCommon"] = d["hidden"] == "0"
        d["comment"] = notes

        for n in ["hidden", "spec_group"]:
            d.pop(n)
        return Specificity(**d)

    def deltad(delta: etree.Element) -> DeltaD:
        return dict(
            mono_mass=float(delta.attrib["mono_mass"]),
            avge_mass=float(delta.attrib["avge_mass"]),
            composition=[
                Composition(
                    **{"symbol": n.attrib["symbol"], "number": int(n.attrib["number"])},
                )
                for n in delta.xpath(".//umod:element", namespaces=root.nsmap)
            ],
        )

    for umod in root.xpath(".//umod:mod", namespaces=root.nsmap):
        a = umod.attrib

        specs = [
            specd(spec)
            for spec in umod.xpath("./umod:specificity", namespaces=root.nsmap)
        ]

        deltas = [
            deltad(delta) for delta in umod.xpath("./umod:delta", namespaces=root.nsmap)
        ]
        assert len(deltas) == 1, nodes(umod)
        delta: DeltaD = deltas[0]

        rec = Unimod(
            unimod_id=int(a["record_id"]),
            name=a["title"],
            approved=bool(a["approved"]),
            description=a["full_name"],
            specificities=specs,  # json.dumps(specs)
            mono_delta=delta["mono_mass"],
            average_delta=delta["avge_mass"],
            composition=delta["composition"],
            # delta=delta,  # used for unimodlookup
        )
        yield rec


def unimodlookup_fromxml(xmlfile: str = "unimod.xml") -> UnimodLookup:
    with open(xmlfile, "rb") as fp:
        x = etree.parse(fp)
    root = x.getroot()
    data = list(parse_unimod(root))
    return UnimodLookup(data)


def make_modcol(
    unimodLookup: UnimodLookup,
    turnover: bool = True,
) -> Callable[[pd.Series], str]:
    # use with df['modcol'] = df.apply(make_modcol(unimodLookup), axis=1)
    delta, mass = ("monoDelta", "monoMass") if turnover else ("massdiff", "mass")

    def m(peptide: str, modifications: list[dict[str, Any]]) -> Iterator[str]:
        for mod in modifications:
            pos = mod["position"]
            ret = unimodLookup.matchModification(peptide, pos, mod[delta])
            if not ret:
                yield f"{mod[mass]:.3f}@{pos}"
            u, _ = ret[0]
            yield f"{u.name}[{pos}]"

    def modcol(row: pd.Series) -> str:
        return ", ".join(m(row["peptide"], row["modifications"]))

    return modcol


if __name__ == "__main__":
    fetch_unimod_from_web()



from __future__ import annotations

from typing import NamedTuple
from typing import TypedDict

import numpy as np

# generated by df2class


class PepXMLRow(NamedTuple):
    run: str
    peptide: str
    calc_neutral_pep_mass: float
    assumed_charge: int
    modcol: str
    precursor_neutral_mass: float
    retention_time_sec: float
    num_missed_cleavages: int
    proteins: np.ndarray  # item: str
    peptideprophet_probability: float
    is_decoy: bool
    agg_count: int
    is_rejected: bool
    mz: float
    observed_mz: float


class PepXMLRunRow(PepXMLRow):
    labelledElementCount: int
    maxIso: int
    mzranges: np.ndarray  # item: float32
    isLabelledRun: bool


class PepXMLRunRowRT(PepXMLRunRow):
    rtmin: float
    rtmax: float


class PepXMLDict(TypedDict):
    run: str
    peptide: str
    calc_neutral_pep_mass: float
    assumed_charge: int
    modcol: str
    observed_mz: float
    precursor_neutral_mass: float
    retention_time_sec: float
    num_missed_cleavages: int
    proteins: np.ndarray  # item: str
    peptideprophet_probability: float
    is_decoy: bool
    agg_count: int
    is_rejected: bool
    mz: float

    labelledElementCount: int
    maxIso: int
    mzranges: np.ndarray  # item: float32
    isLabelledRun: bool


class PepXMLJob(NamedTuple):
    run: str
    peptide: str
    assumed_charge: int
    modcol: str
    calc_neutral_pep_mass: float
    precursor_neutral_mass: float
    mz: float
    observed_mz: float
    rt: float
    num_missed_cleavages: int
    proteins: np.ndarray  # item: str
    peptideprophet_probability: float
    is_decoy: bool
    agg_count: int
    is_rejected: bool
    protein_descr: np.ndarray  # item: str
    modifications: np.ndarray
    prStr: str
    prGroup: int
    labelledElementCount: int
    maxIso: int
    mzranges: np.ndarray  # item: float32
    isLabelledRun: bool



from __future__ import annotations

import logging
from pathlib import Path
from typing import Iterator
from typing import TypeVar

Result = TypeVar("Result")

logger = logging.getLogger("protein_turnover")


def init_logger(
    *,
    level: str | int = logging.INFO,
    logfile: Path | str | None = None,
    reinit: bool = False,
) -> None:
    from .config import LOG_FORMAT

    h: logging.Handler

    if isinstance(level, str):
        level = level.upper()

    def add_handler(h: logging.Handler) -> None:
        fmt = logging.Formatter(LOG_FORMAT)
        h.setFormatter(fmt)
        logger.addHandler(h)

    logger.setLevel(level)
    if reinit:
        for h in list(logger.handlers):
            logger.removeHandler(h)

    if not logger.hasHandlers():
        if logfile is None or logfile == "-":
            add_handler(logging.StreamHandler())
        else:
            add_handler(logging.FileHandler(str(logfile), mode="w"))


def log_iterator(
    it: Iterator[Result],
    *,
    total: int,
    desc: str,
    level: int = 1,
    number_of_bg_processes: int = 1,
) -> Iterator[Result]:
    n = max(1, total // 100)
    if number_of_bg_processes == 1:
        fmt = f"{desc}: [%d/%d %d%%]"
    else:
        fmt = f"{desc}[{level}/{number_of_bg_processes}]: [%d/%d %d%%]"

    for idx, i in enumerate(it, start=1):
        yield i
        if (idx % n) == 0:
            pct = int(idx * 100 / total)
            logger.info(fmt, idx, total, pct)
    logger.info(fmt, total, total, 100)



from __future__ import annotations

import smtplib
from email.mime.text import MIMEText
from typing import Literal


def sendmail(
    html: str,
    you: str,
    *,
    me: str | None = None,
    replyto: str | list[str] | None = None,
    mailhost: str | None = None,
    subject: str | None = None,
    mimetype: Literal["html", "plain"] = "html",
) -> bool:
    from . import config

    if mailhost is None:
        mailhost = config.MAIL_SERVER

    if mailhost is None or mailhost == "none":
        return False

    if subject is None:
        subject = config.MAIL_SUBJECT

    if me is None:
        me = config.MAIL_SENDER

    msg = MIMEText(html, mimetype)

    msg["Subject"] = subject
    msg["From"] = me
    msg["To"] = you
    if replyto is not None:
        msg["Reply-To"] = replyto if isinstance(replyto, str) else ",".join(replyto)

    with smtplib.SMTP() as s:
        s.connect(mailhost)
        s.sendmail(me, [you], msg.as_string())

    return True



from __future__ import annotations

import logging
import mmap
import re
from pathlib import Path
from typing import Iterator
from typing import Literal

import pandas as pd

from .logger import logger
from .utils import human
from .utils import IO
from .utils import ProtXMLResourceFile

PROTEIN = re.compile(b"<protein ")


def scan_proteins(protxml: Path) -> Iterator[int]:
    """Very fast inspection of prot.xml spectra queries"""
    with protxml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in PROTEIN.finditer(mm):
                yield m.start(0)


PROT_OR_PEP = re.compile(
    b'(?P<key>protein_name|unique_stripped_peptides)="(?P<value>[^"]+)"',
)


def full_scan_proteins(
    protxml: Path,
) -> Iterator[
    tuple[Literal["protein_name"], str] | tuple[Literal["peptides"], list[str]]
]:
    """Very fast inspection of prot.xml spectra queries"""
    with protxml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in PROT_OR_PEP.finditer(mm):
                key, value = m.group("key", "value")
                if key == b"protein_name":
                    yield "protein_name", value.decode("ascii")
                else:
                    yield "peptides", value.decode("ascii").split("+")


def full_scan_proteins2(
    protxml: Path,
) -> Iterator[tuple[str, list[str]]]:
    """Very fast inspection of prot.xml spectra queries"""
    ckey: str | None = None
    for key, value in full_scan_proteins(protxml):
        if key == "peptides":
            assert ckey is not None
            yield ckey, value  # type: ignore
        else:
            ckey = value  # type: ignore


def protxml_raw(
    protxml: ProtXMLResourceFile,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> pd.DataFrame:
    from .broken_api import ProtXmlDataFrame

    logger.info("reading: %s", protxml.original.name)
    df = ProtXmlDataFrame(
        protxml.original,
        level=level,
        number_of_bg_processes=number_of_bg_processes,
    )
    logger.info("done: %s", protxml.original.name)
    return df


def getprotxml(protxml: ProtXMLResourceFile) -> pd.DataFrame:
    if protxml.cache_protxml_ok():
        data = protxml.cache_protxml()
        logger.info('getprotxml: reading: "%s"', data.name)
        df = IO(data).read_df()
        logger.info('getprotxml: finished reading: "%s" [%d]', data.name, len(df))

    else:
        if not protxml.exists:
            raise RuntimeError(f"can't find file: {protxml.original}")
        logger.info('getprotxml: creating cache "%s"', protxml.original.name)
        protxml_create(protxml)
        assert protxml.cache_protxml_ok()
        df = IO(protxml.cache_protxml()).read_df()

    return df


def protxml_out(
    df: pd.DataFrame,
    protxml: ProtXMLResourceFile,
) -> None:
    """Write cached version of protxml dataframe"""
    from .utils import IO

    protxmldf = protxml.cache_protxml()

    logger.info("writing file: %s: total=%s", protxmldf.name, len(df))

    df = dehydrate_protxml(df)
    IO(protxmldf, df).save_df()

    if logger.isEnabledFor(logging.INFO):
        mem = df.memory_usage(deep=True)
        size = protxmldf.stat().st_size
        osize = protxml.original.stat().st_size

        logger.info(
            "memory=%s disk=%s original=%s",
            human(mem.sum()),
            human(size),
            human(osize),
        )


def protxml_create(
    protxml: ProtXMLResourceFile,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> int:
    df = protxml_raw(
        protxml,
        level=level,
        number_of_bg_processes=number_of_bg_processes,
    )
    protxml_out(df, protxml)
    return level


def dehydrate_protxml(df: pd.DataFrame) -> pd.DataFrame:
    return df



from __future__ import annotations

import re
import sys
import unicodedata
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import fields
from dataclasses import MISSING
from os.path import commonprefix
from os.path import sep
from pathlib import Path
from typing import Any
from typing import Sequence

import tomli as tomllib
import tomli_w

from .utils import MzMLResourceFile
from .utils import PeptideSettings
from .utils import PepXMLResourceFile
from .utils import ProtXMLResourceFile
from .utils import ResourceFiles

try:
    from unidecode import unidecode
except ImportError:

    def unidecode(string: str, errors: str = "ignore", replace_str: str = "?") -> str:
        return string


WIDE = sys.maxunicode > 0xFFFF


def find_prefix(filenames: Sequence[Path]) -> Path:
    prefix = commonprefix([Path(m).absolute() for m in filenames])
    if not prefix.endswith(sep):
        i = prefix.rfind(sep)
        if i > 0:
            prefix = prefix[: i + 1]
    return Path(prefix)


def jobidkey(pepxml: str, protxml: str, mzmlfiles: list[str]) -> str:
    from hashlib import md5

    m = md5()
    m.update(str(Path(pepxml).resolve()).encode("utf-8"))
    m.update(str(Path(protxml).resolve()).encode("utf-8"))
    for s in mzmlfiles:
        m.update(str(Path(s).resolve()).encode("utf-8"))

    return m.hexdigest()


# def safe_jobid(job_name: str, filename: Path) -> str:

#     jobid = slugify(filename.stem)
#     if len(jobid) >= 5:
#         return jobid
#     if job_name:
#         jobid = slugify(job_name)
#     if not jobid or len(jobid) < 4:
#         from hashlib import md5

#         m = md5()
#         with filename.open("rb") as fp:
#             m.update(fp.read())
#         jobid = m.hexdigest()

#     return jobid


def slugify(s: str, transliterate: bool = True) -> str:
    if not s:
        return s

    if WIDE and transliterate:  # UCS-4 build of python
        s = unidecode(s)
    else:
        s = unicodedata.normalize("NFKD", s)

    slug = s.encode("ascii", "ignore").lower()
    slug = re.sub(b"[^a-z0-9]+", b"-", slug).strip(b"-")
    slug = re.sub(b"[-]+", b"-", slug)
    return slug.decode("ascii")


@dataclass
class TurnoverJob:
    job_name: str
    pepxml: list[str]
    protxml: str
    mzmlfiles: list[str]
    jobid: str = ""
    settings: PeptideSettings = PeptideSettings()
    cache_dir: str | None = None
    email: str | None = None
    status: str | None = None  # pending, started, finished, failed, stopped.
    mzfile_to_run: dict[str, str] | None = None
    match_runNames: bool = False
    auto_generated_jobid: bool = False
    """Match peptides to mzML files via their `spectrum` names"""

    def cache_ok(self) -> bool:
        return len(self.to_resource_files().todo()) == 0

    def get_mzfile_to_run(self) -> dict[str, str]:
        if self.mzfile_to_run is None:
            self.mzfile_to_run = {}
        for mz in self.mzmlfiles:
            pmz = Path(mz)
            if pmz.name not in self.mzfile_to_run:
                self.mzfile_to_run[pmz.name] = pmz.stem
        return self.mzfile_to_run

    @property
    def runNames(self) -> set[str]:
        return set(self.get_mzfile_to_run().values())

    def hash(self) -> str:
        from hashlib import md5

        m = md5()
        m.update(self.settings.hash().encode("utf-8"))
        for pepxml in self.pepxml:
            m.update(str(Path(pepxml).resolve()).encode("utf-8"))
        m.update(str(Path(self.protxml).resolve()).encode("utf-8"))
        for mzml in self.mzmlfiles:
            m.update(str(Path(mzml).resolve()).encode("utf-8"))
        for field in fields(self):
            if field.name in {
                "settings",
                "pepxml",
                "protxml",
                "mzmlfiles",
                "mzfile_to_run",
            }:
                continue
            v = getattr(self, field.name)
            m.update(str(v).encode("utf-8"))
        return m.hexdigest()

    def __hash__(self) -> int:
        return int(self.hash(), 16)

    def save(
        self,
        todir: Path | str | None = None,
        filename: str | None = None,
    ) -> Path:
        outdir = Path(todir or self.cache_dir or ".")
        if not outdir.is_dir():
            outdir.mkdir(exist_ok=True, parents=True)
        if filename is None:
            out = outdir.joinpath(self.jobid + ".toml")
        else:
            if not filename.endswith(".toml"):
                filename += ".toml"
            out = outdir.joinpath(filename)

        return self.save_file(out)

    def save_file(self, out: Path) -> Path:
        mzmlfiles = [Path(m).resolve() for m in self.mzmlfiles]
        prefix = find_prefix(mzmlfiles)

        d = {k: v for k, v in asdict(self).items() if v is not None}

        if "cache_dir" in d:  # might be None
            d["cache_dir"] = str(Path(d["cache_dir"]).expanduser().resolve())

        d["pepxml"] = [str(Path(m).resolve()) for m in self.pepxml]
        d["protxml"] = str(Path(d["protxml"]).resolve())
        d["mzmlfiles"] = [str(f.relative_to(prefix)) for f in mzmlfiles]
        d["mzmlprefix"] = str(Path(prefix).resolve())
        # d["settings"] = asdict(d["settings"])
        # settings = d.pop("settings")
        # d.update(settings)
        if "auto_generated_jobid" in d:
            del d["auto_generated_jobid"]
        with out.open("wb") as fp:
            tomli_w.dump(d, fp)
        return out

    @classmethod
    def safe_jobid(cls, turnover_dict: dict[str, Any], filename: Path) -> str:
        return filename.stem
        # job_name = str(turnover_dict["job_name"])
        # return safe_jobid(job_name, filename)

    @classmethod
    def restore(cls, filename: str | Path) -> TurnoverJob:
        def ensure_list(key: str) -> None:
            if key in turnover_dict:
                r = turnover_dict[key]
                if isinstance(r, str):
                    turnover_dict[key] = [r]

        filename = Path(filename)
        with filename.open("rb") as fp:
            turnover_dict: dict[str, Any] = tomllib.load(fp)
            # just what we want...

            ensure_list("mzmlfiles")
            ensure_list("pepxml")
            if "job_name" not in turnover_dict:
                raise TypeError("please specify a job_name")
            if "jobid" not in turnover_dict:
                turnover_dict["jobid"] = cls.safe_jobid(turnover_dict, filename)
                turnover_dict["auto_generated_jobid"] = True
            if "mzmlprefix" in turnover_dict:
                prefix = Path(turnover_dict.pop("mzmlprefix"))
                turnover_dict["mzmlfiles"] = [
                    str(prefix.joinpath(m)) for m in turnover_dict["mzmlfiles"]
                ]
            if "settings" in turnover_dict:
                settings = turnover_dict["settings"]
                settingsd = {
                    f.name: settings[f.name]
                    for f in fields(PeptideSettings)
                    if f.name in settings
                }
                turnover_dict["settings"] = PeptideSettings(**settingsd)
            else:
                settingsd = {
                    f.name: turnover_dict[f.name]
                    for f in fields(PeptideSettings)
                    if f.name in turnover_dict
                }
                turnover_dict["settings"] = PeptideSettings(**settingsd)
            if "protxml" not in turnover_dict:
                turnover_dict["protxml"] = ""

            missing = {k for k in REQUIRED_FIELDS if k not in turnover_dict}
            if missing:
                s = "" if len(missing) > 1 else ""
                raise ValueError(
                    f"turnover file \"{filename}\" is missing: {', '.join(missing)} value{s}",
                )
            # cleanup
            turnover_dict = {k: v for k, v in turnover_dict.items() if k in ALL_FIELDS}
            return cls(**turnover_dict)

    def verify(self) -> bool:
        from .pymz import verify_run

        return verify_run(self)

    def to_resource_files(self) -> ResourceFiles:
        rf = ResourceFiles(
            [PepXMLResourceFile(m, self.cache_dir) for m in set(self.pepxml)],
            ProtXMLResourceFile(self.protxml, self.cache_dir),
            [MzMLResourceFile(m, self.cache_dir) for m in set(self.mzmlfiles)],
        )
        return rf


ALL_FIELDS = {f.name for f in fields(TurnoverJob)}
REQUIRED_FIELDS = {f.name for f in fields(TurnoverJob) if f.default == MISSING}
